import ast
import re

import numpy
import numpy as np
import pandas

### CATEGORICAL / ONE-HOT EXPANSION ###


def _to_categorical(y, num_classes=None, dtype="float32"):
    """Convert a class vector (integers) to binary class matrix.

    E.g. for use with categorical_crossentropy.

    Parameters
    ----------
    y: class vector to be converted into a matrix
            (integers from 0 to num_classes).
    num_classes: total number of classes.
    dtype: The data type expected by the input, as a string
            (`float32`, `float64`, `int32`...)

    Returns
    -------
    A binary matrix representation of the input. The classes axis
            is placed last.

    Notes
    -----
    This function is from keras.utils
    """
    y = numpy.array(y, dtype="int")
    input_shape = y.shape
    if input_shape and input_shape[-1] == 1 and len(input_shape) > 1:
        input_shape = tuple(input_shape[:-1])
    y = y.ravel()
    if not num_classes:
        num_classes = numpy.max(y) + 1
    n = y.shape[0]
    categorical = numpy.zeros((n, num_classes), dtype=dtype)
    categorical[numpy.arange(n), y] = 1
    output_shape = input_shape + (num_classes,)
    categorical = numpy.reshape(categorical, output_shape)
    return categorical


def to_categorical(arr, index=None, dtype="float32"):
    if isinstance(arr, (pandas.DataFrame, pandas.Series)) and index is None:
        index = arr.index

    codes, positions = numpy.unique(arr, return_inverse=True)
    return pandas.DataFrame(
        data=_to_categorical(positions, dtype=dtype),
        columns=codes,
        index=index,
    )


def categorical_expansion(s, column=None, inplace=False, drop=False):
    """
    Expand a pandas Series into a DataFrame containing a categorical dummy variables.

    This is sometimes called "one-hot" encoding.

    Parameters
    ----------
    s : pandas.Series or pandas.DataFrame
            The input data
    column : str, optional
            If `s` is given as a DataFrame, expand this column
    inplace : bool, default False
            If true, add the result directly as new columns in `s`.
    drop : bool, default False
            If true, drop the existing column from `s`. Has no effect if
            `inplace` is not true.

    Returns
    -------
    pandas.DataFrame
            Only if inplace is false.
    """
    if isinstance(s, pandas.DataFrame):
        input = s
        if column is not None and column not in s.columns:
            raise KeyError(f'key not found "{column}"')
        if len(s.columns) == 1:
            s = s.iloc[:, 0]
            column = s.columns[0]
        elif column is not None:
            s = s.loc[:, column]
    else:
        input = None

    onehot = to_categorical(s)
    onehot.columns = [f"{s.name}=={_}" for _ in onehot.columns]
    if inplace and input is not None:
        input[onehot.columns] = onehot
        if drop:
            input.drop([column], axis=1)
    else:
        return onehot


one_hot_expansion = categorical_expansion


def categorical_compression(df, columns=None, inplace=False, drop=False, name=None):
    """
    Compress a one-hot encoded part of a pandas DataFrame into a categorical variable.

    Parameters
    ----------
    s : pandas.DataFrame
            The input data
    columns : Sequence[str], optional
            Use these columns of `df`, or all columns if not given.
    inplace : bool, default False
            If true, add the result directly as new columns in `df`.
    drop : bool, default False
            If true, drop the existing columns from `df`. Has no effect if
            `inplace` is not true.

    Returns
    -------
    pandas.Series
            Only if inplace is false.
    """
    input = df
    if columns is None:
        columns = df.columns
    onehot = df[columns]
    result = pandas.Series(
        numpy.full(onehot.index.shape, numpy.nan),
        dtype=pandas.CategoricalDtype(columns),
        index=onehot.index,
    )
    for i in onehot.columns:
        result.loc[onehot[i] == 1] = i
    if inplace and name is not None:
        input[name] = result
        if drop:
            input.drop(columns, axis=1)
    else:
        return result


one_hot_compression = categorical_compression


### FOURIER EXPANSION ###


def fourier_expansion_names(basename, length=4):
    """
    Get the names of items for a fourier series expansion.

    Parameters
    ----------
    basename : str
            The input data name
    length : int
            Length of expansion series

    Returns
    -------
    list of str
    """
    columns = []
    for i in range(length):
        func = "cos" if i % 2 else "sin"
        mult = ((i // 2) + 1) * 2
        columns.append(f"{func}({basename}*{mult}*pi)")
    return columns


def fourier_expansion(s, length=4, column=None, inplace=False):
    """
    Expand a pandas Series into a DataFrame containing a fourier series.

    Parameters
    ----------
    s : pandas.Series or pandas.DataFrame
            The input data
    length : int
            Length of expansion series
    column : str, optional
            If `s` is given as a DataFrame, use this column

    Returns
    -------
    pandas.DataFrame
    """
    if isinstance(s, pandas.DataFrame):
        input = s
        if len(s.columns) == 1:
            s = s.iloc[:, 0]
        elif column is not None:
            s = s.loc[:, column]
    else:
        input = None
    columns = fourier_expansion_names(s.name, length=length)
    df = pandas.DataFrame(
        data=0,
        index=s.index,
        columns=columns,
    )
    for i, _col in enumerate(columns):
        func = numpy.cos if i % 2 else numpy.sin
        mult = ((i // 2) + 1) * 2
        df.iloc[:, i] = func(s * mult * numpy.pi)
    if inplace and input is not None:
        input[columns] = df
    else:
        return df


def hard_sigmoid(x, zero_bound, one_bound):
    """
    Apply a piecewise linear sigmoid function.

    Parameters
    ----------
    x : array-like
            Array containing elements to clip
    zero_bound, one_bound : scalar, array-like, or None
            Inflection points of the piecewise linear sigmoid
            function.

    Returns
    -------
    clipped_array : array-like
    """
    x = np.asarray(x)
    if zero_bound < one_bound:
        return np.clip((x - zero_bound) / (one_bound - zero_bound), 0.0, 1.0)
    elif zero_bound == one_bound:
        return (x > zero_bound).astype(float)
    else:
        return np.clip((zero_bound - x) / (zero_bound - one_bound), 0.0, 1.0)


### PIECEWISE LINEAR ###


def piece(x, low_bound, high_bound):
    """
    Clip the values in an array.

    This function differs from the usual `numpy.clip`
    in that the result is shifted by `low_bound` if it is
    given, so that the valid result range is a contiguous
    block of non-negative values starting from 0.

    Parameters
    ----------
    x : array-like
            Array containing elements to clip
    low_bound, high_bound : scalar, array-like, or None
            Minimum and maximum values. If None, clipping is not
            performed on the corresponding edge. Both may be
            None, which results in a noop. Both are broadcast
            against `x`.

    Returns
    -------
    clipped_array : array-like
    """
    if low_bound is None:
        if high_bound is None:
            return x
        else:
            return numpy.minimum(x, high_bound)
    else:
        if high_bound is None:
            return numpy.maximum(x - low_bound, 0)
        else:
            return numpy.clip(x, low_bound, high_bound) - low_bound


def parse_piece(s):
    if s is not None:
        float_regex = r"[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?"
        z = re.match(rf"^piece\((.*),({float_regex}|None),({float_regex}|None)\)$", s)
        if z:
            # data is exactly and only a piecewise part
            base_def = z.group(1)
            lowbound = ast.literal_eval(z.group(2))
            highbound = ast.literal_eval(z.group(4))
            return base_def, lowbound, highbound, True
        z = re.match(rf".*piece\((.*),({float_regex}|None),({float_regex}|None)\)", s)
        if z:
            # data is a piecewise part possibly interacted with something else
            base_def = z.group(1)
            lowbound = ast.literal_eval(z.group(2))
            highbound = ast.literal_eval(z.group(4))
            return base_def, lowbound, highbound, False
    return None, None, None, None


def piecewise_linear(x, p=None, breaks=None):
    from ..jaxmodel import linear
    from ..roles import P, X

    if p is None:
        p = x

    # Flip x and p if given backwards
    if isinstance(x, linear.ParameterRef_C) and isinstance(p, linear.DataRef_C):
        p, x = x, p

    if breaks is None:
        raise ValueError("missing required argument `breaks`")

    xs = piecewise_linear_data_names(x, breaks)
    ps = piecewise_linear_parameter_names(p, breaks)

    return sum(X(x_) * P(p_) for x_, p_ in zip(xs, ps))


def piecewise_linear_parameter_names(basename, breaks):
    from ..roles import P

    width = 1
    if len(breaks) > 9:
        width = 2
    # first leg
    b = breaks[0]
    z = [P(f"{basename}[{1:{width}d}]: up to {b}")]
    # middle legs
    for i in range(len(breaks) - 1):
        b0 = breaks[i]
        b1 = breaks[i + 1]
        z += [P(f"{basename}[{i+2:{width}d}]: {b0} to {b1}")]
    # last leg
    i = len(breaks) - 1
    b0 = breaks[i]
    z += [P(f"{basename}[{i+2:{width}d}]: over {b0}")]
    return z


def piecewise_linear_data_names(basename, breaks):
    # first leg
    b = breaks[0]
    z = [f"piece({basename},None,{b})"]
    # middle legs
    for i in range(len(breaks) - 1):
        b0 = breaks[i]
        b1 = breaks[i + 1]
        z += [f"piece({basename},{b0},{b1})"]
    # last leg
    i = len(breaks) - 1
    b0 = breaks[i]
    z += [f"piece({basename},{b0},None)"]
    return z


def infer_breaks(fun):
    """
    Infer break points from a piecewise_linear LinearFunction.

    Parameters
    ----------
    fun : LinearFunction_C
    """
    m = re.compile(r"piece\(.*,(.*),(.*)\)")
    _0 = m.match(fun[0].data)
    if _0.group(1) != "None":
        raise NotImplementedError("Bounded Minimum")
    breaks = [ast.literal_eval(_0.group(2))]
    for i in fun[1:-1]:
        _n = m.match(i.data)
        if breaks[-1] != ast.literal_eval(_n.group(1)):
            raise NotImplementedError("Diconnected")
        breaks.append(ast.literal_eval(_n.group(2)))
    _n = m.match(fun[-1].data)
    if breaks[-1] != ast.literal_eval(_n.group(1)):
        raise NotImplementedError("Diconnected Top")
    if _n.group(2) != "None":
        raise NotImplementedError("Bounded Maximum")
    return breaks


def piecewise_expansion(s, breaks, column=None, inplace=False):
    """
    Expand a pandas Series into a DataFrame containing a piecewise linear series.

    Parameters
    ----------
    s : pandas.Series or pandas.DataFrame
            The input data
    breaks : list
            Piecewise linear articulation breakpoints
    column : str, optional
            If `s` is given as a DataFrame, use this column

    Returns
    -------
    pandas.DataFrame
    """
    if isinstance(s, pandas.DataFrame):
        input = s
        if len(s.columns) == 1:
            s = s.iloc[:, 0]
        elif column is not None:
            s = s.loc[:, column]
    else:
        input = None

    from ..jaxmodel import linear

    if isinstance(breaks, linear.LinearFunction_C):
        breaks = infer_breaks(breaks)

    columns = piecewise_linear_data_names(s.name, breaks=breaks)
    df = pandas.DataFrame(
        data=0,
        index=s.index,
        columns=columns,
    )

    # first leg
    b = breaks[0]
    df.iloc[:, 0] = piece(s, None, b)

    # middle legs
    for i in range(len(breaks) - 1):
        b0 = breaks[i]
        b1 = breaks[i + 1]
        df.iloc[:, i + 1] = piece(s, b0, b1)
    # last leg
    i = len(breaks) - 1
    b0 = breaks[i]
    df.iloc[:, i + 1] = piece(s, b0, None)

    if inplace and input is not None:
        input[columns] = df
    else:
        return df


### IDCE MANIPULATIONS ###


def _nonzero_minimum_rescale_slower(x):
    x_ = x[x != 0]
    if x_.size == 0:
        s = 1
    else:
        s = numpy.min(x[x != 0])
    return x / s


def _nonzero_minimum_rescale(x):
    s = numpy.ma.masked_equal(x, 0.0, copy=False).min()
    if numpy.isnan(s) or s == 0:
        s = 1
    return x / s


def ratio_min(s, column=None, inplace=False, result_name=None):
    """
    Modify data in `idce` format to scale within cases relative to the minimum value.

    Parameters
    ----------
    s : pandas.Series or pandas.DataFrame
            The input data
    column : str, optional
            If `s` is given as a DataFrame, use this column
    inplace : bool, default False
            modify `s` in-place?
    result_name : str, optional
            If `s` is given as a DataFrame, and `inplace` is True, write to this column.
            When not given, the result is named "ratio_min(<column>)"

    Returns
    -------
    pandas.DataFrame

    Raises
    ------
    ValueError
            The input argument `s` is not in idce format (i.e., it does not have
            a MultiIndex as the index or that MultiIndex does not have 2 levels.)
    """
    index = s.index

    if not isinstance(s.index, pandas.MultiIndex) or len(s.index.levels) != 2:
        raise ValueError("`s` must be in idce format, with a 2-level multi-index.")

    if column is None and isinstance(s, pandas.Series):
        column = s.name

    if isinstance(s, pandas.DataFrame):
        input = s
        if len(s.columns) == 1:
            s = s.iloc[:, 0]
        elif column is not None:
            s = s.loc[:, column]
    else:
        input = None

    divisor = s.groupby(index.codes[0]).transform(numpy.min)
    divisor[divisor == 0] = 1.0
    scaled = s / divisor

    # scaled = s.groupby(index.labels[0]).transform(_nonzero_minimum_rescale)

    if result_name is None and column is not None:
        result_name = f"ratio_min({column})"

    if inplace and input is not None and result_name is not None:
        input[result_name] = scaled
    else:
        return scaled



import os
import sys

import larch


def get_version():
    version = larch.__version__
    os.environ["LARCH_VERSION"] = version
    if len(sys.argv) >= 2:
        v_file = sys.argv[1]
    else:
        v_file = os.path.normpath(
            os.path.join(os.path.dirname(__file__), "..", "LARCH_VERSION.txt")
        )
    with open(v_file, "w") as f:
        f.write(f"LARCH_VERSION={version}\n")
    print(f"LARCH_VERSION={version}")
    return version


if __name__ == "__main__":
    get_version()



import logging
from collections.abc import Mapping

import numba as nb
import numpy as np
import xarray as xr

from .dim_names import ALTID, ALTIDX, CASEALT, CASEID, CASEPTR, GROUPID, INGROUP


@nb.njit
def case_ptr_to_indexes(n_casealts, case_ptrs):
    case_index = np.zeros(n_casealts, dtype=np.int64)
    for c in range(case_ptrs.shape[0] - 1):
        case_index[case_ptrs[c] : case_ptrs[c + 1]] = c
    return case_index


@nb.njit
def ce_dissolve_zero_variance(ce_data, ce_caseptr):
    """
    Dissolves zero variance in the given data based on case pointers.

    Parameters
    ----------
    ce_data : array-like, shape [n_casealts]
        One-dimensional array containing the data.
    ce_caseptr : array-like, shape [n_cases]
        One-dimensional array containing the case pointers.

    Returns
    -------
    out : ndarray
        One-dimensional array containing the dissolved data.
    flag : int
        Flag indicating whether variance was detected or not.
        - 1 if variance was detected
        - 0 if no variance was found and the `out` array is valid.
    """
    failed = 0
    if ce_caseptr.ndim == 2:
        ce_caseptr1 = ce_caseptr[:, -1]
    else:
        ce_caseptr1 = ce_caseptr[1:]
    shape = (ce_caseptr1.shape[0],)
    out = np.zeros(shape, dtype=ce_data.dtype)
    c = 0
    out[0] = ce_data[0]
    for row in range(ce_data.shape[0]):
        if row == ce_caseptr1[c]:
            c += 1
            out[c] = ce_data[row]
        else:
            if out[c] != ce_data[row]:
                failed = 1
                break
    return out, failed


class _GenericFlow:
    def __init__(self, x=None):
        self._obj = x
        self._flow_library = {}

    @property
    def CASEID(self):
        """Str : The _caseid_ dimension of this Dataset, if defined."""
        result = self._obj.attrs.get(CASEID, None)
        return result

    def __set_attr(self, attr_name, target_value, check_dim=True):
        if target_value is None:
            if attr_name in self._obj.attrs:
                del self._obj.attrs[attr_name]
        else:
            if check_dim and target_value not in self._obj.dims:
                raise ValueError(f"cannot set {attr_name}, {target_value} not in dims")
            self._obj.attrs[attr_name] = target_value

    @CASEID.setter
    def CASEID(self, dim_name):
        self.__set_attr(CASEID, dim_name, check_dim=self.GROUPID is not None)

    @property
    def ALTID(self):
        """Str : The _altid_ dimension of this Dataset, if defined."""
        result = self._obj.attrs.get(ALTID, None)
        return result

    @ALTID.setter
    def ALTID(self, dim_name):
        self.__set_attr(ALTID, dim_name)

    @property
    def CASEALT(self):
        """Str : The _casealt_ dimension of this Dataset, if defined."""
        result = self._obj.attrs.get(CASEALT, None)
        return result

    @CASEALT.setter
    def CASEALT(self, dim_name):
        self.__set_attr(CASEALT, dim_name)

    @property
    def ALTIDX(self):
        """Str : The _alt_idx_ dimension of this Dataset, if defined."""
        result = self._obj.attrs.get(ALTIDX, None)
        return result

    @ALTIDX.setter
    def ALTIDX(self, dim_name):
        self.__set_attr(ALTIDX, dim_name, False)

    @property
    def CASEPTR(self):
        """Str : The _caseptr_ dimension of this Dataset, if defined."""
        result = self._obj.attrs.get(CASEPTR, None)
        return result

    @CASEPTR.setter
    def CASEPTR(self, dim_name):
        self.__set_attr(CASEPTR, dim_name, False)

    @property
    def GROUPID(self):
        """Str : The _groupid_ dimension of this Dataset, if defined."""
        result = self._obj.attrs.get(GROUPID, None)
        return result

    @GROUPID.setter
    def GROUPID(self, dim_name):
        self.__set_attr(GROUPID, dim_name)

    @property
    def INGROUP(self):
        """Str : The _ingroup_ dimension of this Dataset, if defined."""
        result = self._obj.attrs.get(INGROUP, None)
        return result

    @INGROUP.setter
    def INGROUP(self, dim_name):
        self.__set_attr(INGROUP, dim_name)

    def set_altids(self, altids, dim_name=None, dtype=None, inplace=False):
        """
        Set the alternative ids for this Dataset.

        Parameters
        ----------
        altids : array-like of int
            Integer id codes.
        dim_name : str, optional
            Use this dimension name for alternatives.
        dtype : dtype, optional
            Coerce the altids to be this dtype.
        inplace : bool, default False
            When true, apply the transformation in-place on the Dataset,
            otherwise return a modified copy.

        Returns
        -------
        Dataset

        Examples
        --------
        >>> import numpy as np
        >>> from larch.dataset import Dataset
        >>> ds = Dataset()
        >>> ds = ds.dc.set_altids([1,2,3,4], dtype=np.int64)
        >>> ds.dc.altids()
        Int64Index([1, 2, 3, 4], dtype='int64', name='_altid_')
        >>> ds.dc.n_alts
        4
        >>> ds2 = ds.dc.set_altids([7,8,9], dim_name='A', dtype=np.int64)
        >>> ds2.dc.ALTID
        'A'
        >>> ds2.dc.altids()
        Int64Index([7, 8, 9], dtype='int64', name='A')
        >>> ds2
        <xarray.Dataset>
        Dimensions:  (_altid_: 4, A: 3)
        Coordinates:
          * _altid_  (_altid_) int64 1 2 3 4
          * A        (A) int64 7 8 9
        Data variables:
            *empty*
        Attributes:
            _altid_:  A
        """
        if inplace:
            obj = self._obj
        else:
            obj = self._obj.copy()
        dim_name = (
            dim_name or getattr(altids, "name", None) or obj.attrs.get(ALTID, ALTID)
        )
        if not isinstance(altids, xr.DataArray):
            altids = xr.DataArray(
                np.asarray(altids),
                dims=(dim_name),
            )
        if dtype is not None:
            altids = altids.astype(dtype)
        obj.coords[dim_name] = altids
        obj.dc.ALTID = dim_name
        return obj

    def set_altnames(self, altnames, inplace=False):
        """
        Set the alternative names for this Dataset.

        Parameters
        ----------
        altnames : Mapping or array-like
            A mapping of (integer) codes to names, or an array or names
            of the same length and order as the alternatives already
            defined in this Dataset.
        inplace : bool, default False
            When true, apply the transformation in-place on the Dataset,
            otherwise return a modified copy.

        Returns
        -------
        Dataset
        """
        if inplace:
            obj = self._obj
        else:
            obj = self._obj.copy()
        if isinstance(altnames, Mapping):
            a = obj.dc.ALTID
            names = xr.DataArray(
                [altnames.get(i, None) for i in obj[a].values],
                dims=a,
            )
        elif isinstance(altnames, xr.DataArray):
            names = altnames
        else:
            if obj.dc.ALTID is None:
                obj = obj.dc.set_altids(np.arange(1, len(altnames) + 1))
            names = xr.DataArray(
                np.asarray(altnames),
                dims=obj.dc.ALTID,
            )
        obj.coords["alt_names"] = names
        return obj

    def as_tree(self, label="main", exclude_dims=()):
        """
        Convert this Dataset to a DataTree.

        For |idco| and |idca| datasets, the result will generally be a
        single-node tree.  For |idce| data, there will be

        Parameters
        ----------
        label : str, default 'main'
            Name to use for the root node in the tree.
        exclude_dims : Tuple[str], optional
            Exclude these dimensions, in addition to any
            dimensions listed in the `_exclude_dims_` attribute.

        Returns
        -------
        DataTree
        """
        from ..dataset import DataTree

        if self.CASEPTR is not None:
            case_index = case_ptr_to_indexes(
                self._obj.sizes[self.CASEALT], self[self.CASEPTR].values
            )
            obj = self._obj.assign(
                {"_case_index_": xr.DataArray(case_index, dims=(self.CASEALT))}
            )
            tree = DataTree(**{label: obj.drop_dims(self.CASEID)})
            ds = obj.keep_dims(self.CASEID)
            ds.attrs.pop("_exclude_dims_", None)
            ds.attrs.pop("_caseptr_", None)
            ds.attrs.pop("_casealt_", None)
            ds.attrs.pop("_alt_idx_", None)
            tree.add_dataset(
                "idcoVars",
                ds,
                relationships=(f"{label}._case_index_ -> idcoVars.{self.CASEID}"),
            )
        else:
            tree = DataTree(**{label: self._obj})
        return tree

    def setup_flow(self, *args, **kwargs):
        """
        Set up a new Flow for analysis using the structure of this DataTree.

        This method creates a new `DataTree` with only this Dataset as
        the root Dataset labeled `main`.  All other arguments are passed
        through to `DataTree.setup_flow`.

        Returns
        -------
        Flow
        """
        return self.as_tree().setup_flow(*args, **kwargs)

    def caseids(self):
        """
        Access the caseids coordinates as an index.

        Returns
        -------
        pd.Index
        """
        return self.indexes[self.CASEID]

    def altids(self):
        """
        Access the altids coordinates as an index.

        Returns
        -------
        pd.Index
        """
        return self._obj.indexes[self.ALTID]

    def groupids(self):
        """
        Access the groupids coordinates as an index.

        Returns
        -------
        pd.Index
        """
        return self.indexes[self.GROUPID]

    @property
    def alts_mapping(self):
        """Dict[int,str] : Mapping of alternative codes to names."""
        a = self._obj.coords[self.ALTID]
        if "alt_names" in a.coords:
            return dict(zip(a.values, a.coords["alt_names"].values))
        else:
            return dict(zip(a.values, a.values))

    @property
    def n_cases(self):
        try:
            return self.sizes[self.CASEID]
        except KeyError:
            try:
                return self.sizes[self.GROUPID] * self.sizes[self.INGROUP]
            except KeyError:
                pass
            logging.getLogger().error(
                f"missing {self.CASEID!r} among dims {self.sizes}"
            )
            raise

    @property
    def n_panels(self):
        try:
            return self.sizes[self.GROUPID]
        except KeyError:
            try:
                return self.sizes[self.CASEID]
            except KeyError:
                pass
            logging.getLogger().error(
                f"missing {self.GROUPID!r} and {self.CASEID!r} among dims {self.sizes}"
            )
            raise

    @property
    def n_in_panel(self):
        return self.sizes[self.INGROUP]

    def transfer_dimension_attrs(self, target):
        if not isinstance(target, xr.DataArray | xr.Dataset):
            return target
        updates = {}
        for i in [CASEID, ALTID, GROUPID, INGROUP, CASEALT, CASEPTR]:
            j = self._obj.attrs.get(i, None)
            if j is not None and j in target.dims:
                updates[i] = j
        return target.assign_attrs(updates)

    def get_expr(self, expression):
        """
        Access or evaluate an expression.

        Parameters
        ----------
        expression : str

        Returns
        -------
        DataArray
        """
        try:
            result = self._obj[expression]
        except (KeyError, IndexError):
            if expression in self._flow_library:
                flow = self._flow_library[expression]
            else:
                flow = self.setup_flow({expression: expression})
                self._flow_library[expression] = flow
            if flow.tree.root_dataset is not self:
                flow.tree = self.as_tree()
            result = flow.load_dataarray().isel(expressions=0)
        result = self.transfer_dimension_attrs(result)
        return result

    def dissolve_zero_variance(self, dim="<ALTID>", inplace=False):
        """
        Dissolve dimension on variables where it has no variance.

        This method is convenient to convert variables that have
        been loaded as |idca| or |idce| format into |idco| format where
        appropriate.

        Parameters
        ----------
        dim : str, optional
            The name of the dimension to potentially dissolve.
        inplace : bool, default False
            Whether to dissolve variables in-place.

        Returns
        -------
        Dataset
        """
        if dim == "<ALTID>":
            dim = self.ALTID
        if inplace:
            obj = self
        else:
            obj = self.copy()
        for k in obj.variables:
            if obj[k].dtype.kind in {"U", "S", "O"}:
                continue
            if dim in obj[k].dims:
                try:
                    dissolve = obj[k].std(dim=dim).max() < 1e-10
                except TypeError:
                    pass
                else:
                    if dissolve:
                        obj[k] = obj[k].min(dim=dim)
            elif obj[k].dims == (self.CASEALT,):
                proposal, flag = ce_dissolve_zero_variance(
                    obj[k].values, obj[self.CASEPTR].values
                )
                if flag == 0:
                    obj = obj.assign({k: xr.DataArray(proposal, dims=(self.CASEID))})
        return obj

    def dissolve_coords(self, dim, others=None):
        d = self._obj.reset_index(dim)
        a = d[f"{dim}_"]
        mapper = dict((j, i) for (i, j) in enumerate(a.to_series()))
        mapper_f = np.vectorize(mapper.get)
        if others is None:
            others = []
        if isinstance(others, str):
            others = [others]
        for other in others:
            d[other] = xr.apply_ufunc(mapper_f, d[other])
        return d


@xr.register_dataarray_accessor("dc")
class _DataArrayDC(_GenericFlow):
    _parent_class = xr.DataArray

    @property
    def n_alts(self):
        if self.ALTID in self._obj.sizes:
            return self._obj.shape[self.sizes.index(self.ALTID)]
        if "n_alts" in self._obj.attrs:
            return self._obj.attrs["n_alts"]
        raise ValueError("no n_alts set")

    def __getitem__(self, name):
        # pass dimension attrs to DataArray
        result = self._obj[name]
        result = self.transfer_dimension_attrs(result)
        return result

    def __getattr__(self, name):
        # pass dimension attrs to DataArray
        result = getattr(self._obj, name)
        result = self.transfer_dimension_attrs(result)
        return result


@xr.register_dataset_accessor("dc")
class _DatasetDC(_GenericFlow):
    _parent_class = xr.Dataset

    @property
    def n_alts(self):
        if self.ALTID in self._obj.sizes:
            return self._obj.sizes[self.ALTID]
        if "n_alts" in self._obj.attrs:
            return self._obj.attrs["n_alts"]
        raise ValueError("no n_alts set")

    def __getitem__(self, name):
        # pass dimension attrs to DataArray
        result = self._obj[name]
        result = self.transfer_dimension_attrs(result)
        return result

    def __getattr__(self, name):
        # pass dimension attrs to DataArray
        result = getattr(self._obj, name)
        result = self.transfer_dimension_attrs(result)
        return result

    def __contains__(self, item):
        return self._obj.__contains__(item)

    def query_cases(self, query, parser="pandas", engine=None):
        """
        Return a new dataset with each array indexed along the CASEID dimension.

        The indexers are given as strings containing Python expressions to be
        evaluated against the data variables in the dataset.

        Parameters
        ----------
        query : str
            Python expressions to be evaluated against the data variables
            in the dataset. The expressions will be evaluated using the pandas
            eval() function, and can contain any valid Python expressions but cannot
            contain any Python statements.
        parser : {"pandas", "python"}, default: "pandas"
            The parser to use to construct the syntax tree from the expression.
            The default of 'pandas' parses code slightly different than standard
            Python. Alternatively, you can parse an expression using the 'python'
            parser to retain strict Python semantics.
        engine : {"python", "numexpr", None}, default: None
            The engine used to evaluate the expression. Supported engines are:

            - None: tries to use numexpr, falls back to python
            - "numexpr": evaluates expressions using numexpr
            - "python": performs operations as if you had eval’d in top level python

        Returns
        -------
        obj : Dataset
            A new Dataset with the same contents as this dataset, except each
            array is indexed by the results of the query on the CASEID dimension.

        See Also
        --------
        Dataset.isel
        pandas.eval
        """
        result = self._obj.query({self.CASEID: query}, parser=parser, engine=engine)
        result = self.transfer_dimension_attrs(result)
        return result

    def to_arrays(self, graph, float_dtype=np.float64):
        from ..model.cascading import array_av_cascade, array_ch_cascade
        from ..model.data_arrays import DataArrays

        if "co" in self:
            co = self["co"].values.astype(float_dtype)
        else:
            co = np.empty((self.n_cases, 0), dtype=float_dtype)

        if "ca" in self:
            ca = self["ca"].values.astype(float_dtype)
        else:
            ca = np.empty((self.n_cases, self.n_alts, 0), dtype=float_dtype)

        if "ce_data" in self:
            ce_data = self["ce_data"].values.astype(float_dtype)
        else:
            ce_data = np.empty((0, 0), dtype=float_dtype)

        if self.ALTIDX is not None:
            ce_altidx = self[self.ALTIDX].values
        else:
            ce_altidx = np.empty((0), dtype=np.int16)

        if self.CASEPTR is not None:
            ce_caseptr = np.lib.stride_tricks.sliding_window_view(
                self[self.CASEPTR].values, 2
            )
        else:
            ce_caseptr = np.empty((self.n_cases, 0), dtype=np.int16)

        if "wt" in self:
            wt = self["wt"].values.astype(float_dtype)
        else:
            if self.GROUPID is None:
                wt = np.ones(self.n_cases, dtype=float_dtype)
            else:
                wt = np.ones(self.n_panels, dtype=float_dtype)

        if "ch" in self:
            ch = array_ch_cascade(self["ch"].values, graph, dtype=float_dtype)
        else:
            if self.GROUPID is None:
                ch = np.zeros([self.n_cases, len(graph)], dtype=float_dtype)
            else:
                ch = np.zeros(
                    [self.n_panels, self.n_in_panel, len(graph)], dtype=float_dtype
                )

        if "av" in self:
            av = array_av_cascade(self["av"].values, graph)
        else:
            if self.GROUPID is None:
                av = np.ones([self.n_cases, len(graph)], dtype=np.int8)
            else:
                av = np.ones(
                    [self.n_panels, self.n_in_panel, len(graph)], dtype=np.int8
                )

        return DataArrays(ch, av, wt, co, ca, ce_data, ce_altidx, ce_caseptr)


@xr.register_dataset_accessor("icase")
class _DatasetCaseIslice(_GenericFlow):
    def __getitem__(self, item):
        if self.CASEID is not None:
            return self._obj.isel({self.CASEID: item})
        elif self.GROUPID is not None:
            return self._obj.isel({self.GROUPID: item})
        else:
            raise ValueError("neither CASEID nor GROUPID is defined")



import numpy
import pandas

from ..shorts import LinearFunction, P, X

# duplicate these here for legacy compatibility


def polynomial(x, p, powers=None, funs=None):
    if powers is None:
        powers = {
            1: "",
            2: "_Squared",
            3: "_Cubed",
        }
    z = LinearFunction()
    for pwr, label in powers.items():
        if pwr == 1:
            z = z + X(x) * P(f"{p}{label}")
        else:
            z = z + X(f"({x})**{pwr}") * P(f"{p}{label}")

    if funs is not None:
        for fun, label in funs.items():
            z = z + X(f"{fun}({x})") * P(f"{p}{label}")

    return z


def fourier_series(x, p=None, length=4):
    z = LinearFunction()
    if p is None:
        p = x
    for i in range(length):
        func = "cos" if i % 2 else "sin"
        mult = ((i // 2) + 1) * 2
        z = z + X(f"{func}({x}*{mult}*pi)") * P(f"{func}_{mult}π{p}")
    return z


def fourier_expansion_names(basename, length=4):
    """
    Get the names of items for a fourier series expansion.

    Parameters
    ----------
    basename : str
        The input data name
    length : int
        Length of expansion series

    Returns
    -------
    list of str
    """
    columns = []
    for i in range(length):
        func = "cos" if i % 2 else "sin"
        mult = ((i // 2) + 1) * 2
        columns.append(f"{func}({basename}*{mult}*pi)")
    return columns


def fourier_expansion(s, length=4, column=None, inplace=False):
    """
    Expand a pandas Series into a DataFrame containing a fourier series.

    Parameters
    ----------
    s : pandas.Series or pandas.DataFrame
        The input data
    length : int
        Length of expansion series
    column : str, optional
        If `s` is given as a DataFrame, use this column

    Returns
    -------
    pandas.DataFrame
    """
    if isinstance(s, pandas.DataFrame):
        input = s
        if len(s.columns) == 1:
            s = s.iloc[:, 0]
        elif column is not None:
            s = s.loc[:, column]
    else:
        input = None
    columns = fourier_expansion_names(s.name, length=length)
    df = pandas.DataFrame(
        data=0,
        index=s.index,
        columns=columns,
    )
    for i, _col in enumerate(columns):
        func = numpy.cos if i % 2 else numpy.sin
        mult = ((i // 2) + 1) * 2
        df.iloc[:, i] = func(s * mult * numpy.pi)
    if inplace and input is not None:
        input[columns] = df
    else:
        return df


def normalize(x, std_div=1):
    """
    Normalize an array by subtracting the mean and dividing by the standard deviation.

    Parameters
    ----------
    x : ndarray
    std_div : numeric, optional
        Divide by this many standard deviations.  Defaults to 1, but you may consider using 2
        per [1]_.

    Returns
    -------
    ndarray

    References
    ----------
    .. [1] A. Gelman, Scaling regression inputs by dividing by two standard deviations,
       Statistics in medicine 27 (15) (2008) 2865–2873.
    """
    m = numpy.nanmean(x)
    s = numpy.nanstd(x)
    s *= std_div
    return (x - m) / s


def boolean(x):
    """
    Make an array boolean.

    Parameters
    ----------
    x : array-like

    Returns
    -------
    array-like
    """
    try:
        return x.astype(bool)
    except Exception:
        return numpy.asarray(x, dtype="bool")



import numpy as np
import pandas as pd
import pytest
from pytest import approx

import larch as lx
from larch import P, X


@pytest.mark.parametrize(
    "compute_engine,use_streaming",
    [
        ("jax", False),
        ("numba", False),
        # ("numba", True),
    ],
)
def test_destination_only(compute_engine, use_streaming):
    hh, pp, tour, skims, emp = lx.example(200, ["hh", "pp", "tour", "skims", "emp"])
    hh["INCOME_GRP"] = pd.qcut(hh.INCOME, 3)
    co = lx.Dataset.construct(
        tour.set_index("TOURID"),
        caseid="TOURID",
        alts=skims.TAZ_ID,
    )
    tree = lx.DataTree(
        base=co,
        hh=hh.set_index("HHID"),
        person=pp.set_index("PERSONID"),
        emp=emp,
        skims=lx.Dataset.construct.from_omx(skims),
        relationships=(
            "base.TAZ_ID @ emp.TAZ",
            "base.HHID @ hh.HHID",
            "base.PERSONID @ person.PERSONID",
            "hh.HOMETAZ @ skims.otaz",
            "base.TAZ_ID @ skims.dtaz",
        ),
    )
    m = lx.Model(
        datatree=tree, compute_engine=compute_engine, use_streaming=use_streaming
    )
    m.title = "Exampville Tour Destination Choice v2"
    m.quantity_ca = (
        +P.EmpRetail_HighInc * X("RETAIL_EMP * (INCOME>50000)")
        + P.EmpNonRetail_HighInc * X("NONRETAIL_EMP") * X("INCOME>50000")
        + P.EmpRetail_LowInc * X("RETAIL_EMP") * X("INCOME<=50000")
        + P.EmpNonRetail_LowInc * X("NONRETAIL_EMP") * X("INCOME<=50000")
    )
    m.quantity_scale = P.Theta
    m.utility_ca = +P.distance * X.AUTO_DIST
    m.choice_co_code = "base.DTAZ"
    m.plock(EmpRetail_HighInc=0, EmpRetail_LowInc=0)
    assert m.loglike() == approx(-77777.17321427424)
    assert m.d_loglike() == approx(
        [-223.950165, -682.110256, 0.0, 0.0, -7406.389922, -34762.91256], rel=1e-2
    )

    result = m.maximize_loglike(stderr=True)
    assert result.loglike == approx(-70650.07578452416)
    assert result.success
    assert result.n_cases == 20739
    assert result.logloss == approx(3.4066288531040145)
    x = result.x
    if not isinstance(x, pd.Series):
        x = pd.Series(x, index=m.pnames)
    pd.testing.assert_series_equal(
        x.sort_index(),
        pd.Series(
            {
                "EmpNonRetail_HighInc": 1.2453335020460703,
                "EmpNonRetail_LowInc": -1.0893594261458912,
                "EmpRetail_HighInc": 0.0,
                "EmpRetail_LowInc": 0.0,
                "Theta": 0.676440163641688,
                "distance": -0.3347118435209836,
            }
        ).sort_index(),
        rtol=1e-2,
    )
    assert m.pstderr == approx(
        np.array([0.145749, 0.052355, 0.0, 0.0, 0.009012, 0.003812]),
        rel=1e-2,
    )



import base64
import io

from PIL import Image
from xmle import Elem


def make_png(
    content,
    dpi=None,
    compress=True,
    output="Elem",
    close_after=True,
    pad_inches=0.1,
    facecolor=None,
    return_size=False,
    return_dpi=False,
):
    _bytes_io = None
    _img = None
    _size = None

    if hasattr(content, "_repr_png_"):
        content = content._repr_png_()

    if isinstance(content, bytes):
        _bytes_io = io.BytesIO(content)

    if "matplotlib" in str(type(content)):
        import matplotlib.figure

        from .._optional import pyplot as plt

        if not isinstance(content, matplotlib.figure.Figure):
            try:
                content = content.get_figure()
            except AttributeError:
                if not hasattr(content, "savefig"):
                    raise TypeError(
                        "matplotlib content must provide `get_figure` or `savefig` method."
                    ) from None

        # content is a Figure or otherwise has a savefig method
        try:
            fig_number = content.number
        except AttributeError:
            fig_number = None

        if facecolor is None:
            try:
                facecolor = content.get_facecolor()
            except Exception:
                facecolor = "w"

        try:
            edgecolor = content.get_edgecolor()
        except Exception:
            edgecolor = "none"

        _bytes_io = io.BytesIO()
        content.savefig(
            _bytes_io,
            dpi=dpi,
            orientation="portrait",
            format="png",
            bbox_inches="tight",
            pad_inches=pad_inches,
            facecolor=facecolor,
            edgecolor=edgecolor,
        )

        if close_after and fig_number is not None:
            plt.close(fig_number)

    if _bytes_io.getvalue()[:4] == b"\x89PNG":
        _img = Image.open(_bytes_io)
        dpi = _img.info.get("dpi", dpi)
        if dpi is None:
            dpi = [96, 96]
        if compress:
            _img = _img.convert(mode="P", palette="ADAPTIVE")
        _bytes_io = io.BytesIO()
        _img.save(_bytes_io, dpi=dpi, format="png")
        _size = _img.size
    else:
        raise ValueError("Not valid PNG data")

    if output.lower() == "elem":
        result = Elem(
            tag="img",
            src=f"data:image/png;base64,{base64.standard_b64encode(_bytes_io.getvalue()).decode()}",
        )
    elif output.lower() == "bytesio":
        result = _bytes_io
    else:
        result = _img

    if return_size:
        if isinstance(result, tuple):
            result = (*result, _size)
        else:
            result = result, _size

    if return_dpi:
        if isinstance(result, tuple):
            result = (*result, dpi)
        else:
            result = result, dpi

    return result



def example(extract='m', estimate=False):
    import pandas as pd

    import larch as lx

    raw_data = pd.read_csv(lx.example_file("swissmetro.csv.gz")).rename_axis(index="CASEID")
    data = lx.Dataset.construct.from_idco(raw_data, alts={1: "Train", 2: "SM", 3: "Car"})
    m = lx.Model(data.dc.query_cases("PURPOSE in (1,3) and CHOICE != 0"))

    m.title = "swissmetro example 09 (nested logit)"

    m.availability_co_vars = {
        1: "TRAIN_AV * (SP!=0)",
        2: "SM_AV",
        3: "CAR_AV * (SP!=0)",
    }
    m.choice_co_code = "CHOICE"

    from larch import P, X

    m.utility_co[1] = P.ASC_TRAIN + P.B_TIME * X.TRAIN_TT + P.B_COST * X("TRAIN_CO*(GA==0)")
    m.utility_co[2] = P.B_TIME * X.SM_TT + P.B_COST * X("SM_CO*(GA==0)")
    m.utility_co[3] = P.ASC_CAR + P.B_TIME * X.CAR_TT + P.B_COST * X("CAR_CO")

    m.graph.new_node(parameter="existing", children=[1, 3], name="Existing")

    m.graph

    m.ordering = [
        (
            "ASCs",
            "ASC.*",
        ),
        (
            "LOS",
            "B_.*",
        ),
        ("LogSums", "existing"),
    ]
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]
    m.set_cap(15)
    result = m.maximize_loglike()
    result

    m.calculate_parameter_covariance(robust=True)
    m.parameter_summary()
    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



from .interface_info import ipython_status as _ipython_status
from .rate_limiter import NonBlockingRateLimiter

_message_set = _ipython_status()


class Throttlable:
    def __init__(self, throttle=None):
        if throttle is not None:
            self._throttle = NonBlockingRateLimiter(throttle)
        else:
            self._throttle = True


if "IPython" in _message_set:
    from IPython.display import HTML, clear_output, display, display_html

    class display_head(Throttlable):
        def __init__(self, text, level=3, throttle=2):
            super().__init__(throttle)
            self.level = level
            self.tag = display(
                HTML(f"<h{self.level}>{text}</h{self.level}>"), display_id=True
            )

        def update(self, text, newline=False, force=False):
            if self._throttle or force:
                self.tag.update(HTML(f"<h{self.level}>{text}</h{self.level}>"))

        def __call__(self, *text, **kwargs):
            self.update(*text, **kwargs)

        def linefeed(self):
            pass

    class display_p(Throttlable):
        def __init__(self, text, force=False, throttle=2):
            super().__init__(throttle)
            if isinstance(text, str):
                self.tag = display(HTML(f"<p>{text}</p>"), display_id=True)
            else:
                self.tag = display(text, display_id=True)

        def update(self, text, force=False):
            if self._throttle or force:
                if isinstance(text, str):
                    self.tag.update(HTML(f"<p>{text}</p>"))
                else:
                    self.tag.update(text)

        def __call__(self, *text, **kwargs):
            self.update(*text, **kwargs)

else:

    class fake_display:
        def __init__(self, *args, **kwargs):
            print(*args)

        def update(self, *args):
            print(*args)

    display = fake_display
    clear_output = lambda *x, **y: None
    display_html = lambda *x, **y: None
    HTML = lambda *x, **y: None

    class display_head(Throttlable):
        def __init__(self, text, level=3, throttle=2):
            super().__init__(throttle)
            self.level = level
            print(f"{text}: ", end="")

        def update(self, text, newline=False, force=False):
            if self._throttle or force:
                if newline:
                    print("")
                    print(f"{text} ", end="")
                else:
                    print(f"{text}: ", end="")

        def __call__(self, *text, **kwargs):
            self.update(*text, **kwargs)

        def linefeed(self):
            print("")

    class display_p(Throttlable):
        def __init__(self, text, force=False, throttle=2):
            super().__init__(throttle)
            if isinstance(text, str):
                print(text)
            else:
                if force:
                    print(str(text))

        def update(self, text, force=False):
            if self._throttle or force:
                if isinstance(text, str):
                    print(text)
                else:
                    if force:
                        print(str(text))

        def __call__(self, *text, **kwargs):
            self.update(*text, **kwargs)


class display_nothing:
    def __init__(self, *args, **kwargs):
        pass

    def update(self, *args, **kwargs):
        pass

    def __call__(self, *text, **kwargs):
        pass

    def linefeed(self):
        pass



import base64
import logging
import time

import matplotlib.axes
import matplotlib.figure
import numpy as np
import pandas as pd
from pandas.io.formats.style import Styler
from xmle import Elem

from .. import Model, __version__
from ..jaxmodel.model_group import ModelGroup
from .png import make_png

logger = logging.getLogger("Larch.Excel")

try:
    import xlsxwriter
except ImportError:
    xlsxwriter = None
    _XlsxWriter = object
else:
    try:
        from pandas.io.excel._xlsxwriter import _XlsxWriter
    except ImportError:
        from pandas.io.excel import _XlsxWriter


class NumberedCaptions:
    def __init__(self, kind):
        self._kind = kind
        self._n = 1

    def __call__(self, caption):
        if caption:
            numb_caption = f"{self._kind} {self._n}: {caption}"
            self._n += 1
            return numb_caption
        else:
            return caption


class ExcelWriter(_XlsxWriter):
    engine = "xlsxwriter_larch"
    supported_extensions = (".xlsx",)

    def __init__(
        self,
        *args,
        model=None,
        data_statistics=True,
        nesting=True,
        numbering=True,
        hide_log=True,
        output_renderer=None,
        embed_model=True,
        engine_kwargs=None,
        **kwargs,
    ):
        _engine = "xlsxwriter_larch"
        kwargs.pop("engine", None)
        options = kwargs.pop("workbook_options", {})
        if "nan_inf_to_errors" not in options:
            options["nan_inf_to_errors"] = True
        kwargs["options"] = options
        super().__init__(*args, engine=_engine, engine_kwargs=kwargs)
        self.book.set_size(1600, 1200)
        self.head_fmt = self.book.add_format({"bold": True, "font_size": 14})
        self.ital_fmt = self.book.add_format({"italic": True, "font_size": 12})
        self.toc_link_fmt = self.book.add_format({"font_size": 8})

        self.fixed_precision = {2: "_-0.00;-0.00"}

        self.sheet_startrow = {}
        self._col_widths = {}
        if output_renderer is None:
            output_renderer = lambda x: x
        self._output_renderer = output_renderer

        if numbering:
            self.TAB = NumberedCaptions("Table")
            self.FIG = NumberedCaptions("Figure")
        else:
            self.TAB = lambda x: x
            self.FIG = lambda x: x

        self.tocsheet = self.add_worksheet("Contents")  # first sheet cannot be hidden
        if model is not None:
            self.tocsheet.write(0, 0, model.title, self.head_fmt)
        self.tocsheet.write(2, 0, f"   🌳 Larch {__version__}")
        self.tocsheet.write(3, 0, time.strftime("     %A %d %B %Y, %I:%M:%S %p %Z"))
        self.tocsheet.write(5, 0, "Table of Contents", self.head_fmt)
        self.sheet_startrow["Contents"] = 6

        self.logsheet = self.add_worksheet("_log_", hide=hide_log)
        self.log(f"larch.util.excel.ExcelWriter opened: {str(args)}")

        self.metadatasheet = self.add_worksheet("_metadata_", hide=True)
        self.metadatasheet.write(0, 0, "key")
        self.metadatasheet.write(0, 1, "len")
        self.metadatasheet.write(0, 2, "val")
        self.sheet_startrow["_metadata_"] = 1

        if model is not None:
            self.add_model(
                model,
                data_statistics=data_statistics,
                nesting=nesting,
                embed=embed_model,
            )

    def add_model(
        self,
        model,
        data_statistics=True,
        nesting=True,
        utility_functions=True,
        on_error="pass",
        embed=True,
    ):
        ps_data_index_nlevels = 1
        try:
            ps = model.parameter_summary("df")
            ps_data_index_nlevels = ps.data.index.nlevels
            if "Constrained" in ps.data.columns:
                ps.data["Constrained"] = ps.data["Constrained"].str.replace(
                    "<br/>", ",\n"
                )
            self.add_content_tab(ps, sheetname="Parameters", heading="Parameters")
            if "Constrained" in ps.data.columns:
                n = ps.data.index.nlevels + ps.data.columns.get_loc("Constrained")
                if "Parameters" not in self._col_widths:
                    self._col_widths["Parameters"] = {}
                current_width = self._col_widths["Parameters"].get(n, 8)
                new_width = max(current_width, 50)
                self._col_widths["Parameters"][n] = new_width
                wrap = self.book.add_format()
                wrap.set_text_wrap()
                self.sheets["Parameters"].set_column(
                    n, n, width=new_width, cell_format=wrap
                )
        except Exception:
            if on_error == "raise":
                raise

        try:
            startrow = self.sheet_startrow.get("Parameters", 0)
            worksheet = self.sheets["Parameters"]
            worksheet.write_url(
                startrow,
                0,
                "internal:Contents!A1",
                self.toc_link_fmt,
                string="<< Back to Table of Contents",
            )
            startrow += 1
            h = self.TAB("Estimation Statistics")
            worksheet.write(startrow, 0, h, self.head_fmt)
            self._add_toc_entry(h, "Parameters", startrow)
            startrow += 1
            logger.debug(" after heading row is %d", startrow)
            startrow = _estimation_statistics_excel(
                model,
                self,
                "Parameters",
                start_row=startrow,
                buffer_cols=ps_data_index_nlevels - 1,
            )
        except Exception:
            if on_error == "raise":
                raise

        if data_statistics and not isinstance(model, ModelGroup):
            from .statistics import statistics_for_dataframe

            if model.dataframes is not None:
                if model.dataframes.data_co is not None and len(
                    model.dataframes.data_co.columns
                ):
                    try:
                        self.add_content_tab(
                            statistics_for_dataframe(model.dataframes.data_co),
                            sheetname="CO Data",
                            heading="CO Data",
                        )
                    except Exception:
                        if on_error == "raise":
                            raise
                if model.dataframes.data_ca is not None and len(
                    model.dataframes.data_ca.columns
                ):
                    try:
                        self.add_content_tab(
                            statistics_for_dataframe(model.dataframes.data_ca),
                            sheetname="CA Data",
                            heading="CA Data",
                        )
                    except Exception:
                        if on_error == "raise":
                            raise
                if model.dataframes.data_ce is not None and len(
                    model.dataframes.data_ce.columns
                ):
                    try:
                        self.add_content_tab(
                            statistics_for_dataframe(model.dataframes.data_ce),
                            sheetname="CE Data",
                            heading="CE Data",
                        )
                    except Exception:
                        if on_error == "raise":
                            raise
                if (
                    model.dataframes.data_ch is not None
                    and model.dataframes.data_av is not None
                ):
                    try:
                        self.add_content_tab(
                            model.dataframes.choice_avail_summary(
                                graph=model.graph,
                                availability_co_vars=model.availability_co_vars,
                            ),
                            sheetname="Choice",
                            heading="Choices",
                        )
                    except Exception:
                        if on_error == "raise":
                            raise

        if utility_functions:
            try:
                if isinstance(model, ModelGroup):
                    for m in model:
                        self.add_content_tab(
                            m._utility_functions_as_frame(),
                            sheetname="Utility",
                            heading=f"Utility Functions {m.title}",
                        )
                else:
                    self.add_content_tab(
                        model._utility_functions_as_frame(),
                        sheetname="Utility",
                        heading="Utility Functions",
                    )
                self.sheets["Utility"].set_column("B:B", None, None, {"hidden": 1})
            except Exception:
                if on_error == "raise":
                    raise

        if nesting and not isinstance(model, ModelGroup) and not model.is_mnl():
            try:
                self.add_content_tab(
                    model.graph.__xml__(output="png"),
                    sheetname="Nesting",
                    heading="Nesting Tree",
                )
            except Exception:
                if on_error == "raise":
                    raise
            try:
                self.add_content_tab(
                    model.graph_descrip("nodes"),
                    sheetname="Nesting",
                    heading="Nesting Node List",
                )
            except Exception:
                if on_error == "raise":
                    raise

        elif nesting and isinstance(model, ModelGroup):
            for m in model:
                if not m.is_mnl():
                    try:
                        self.add_content_tab(
                            m.graph.__xml__(output="png"),
                            sheetname="Nesting",
                            heading="Nesting Tree",
                        )
                    except Exception:
                        if on_error == "raise":
                            raise
                    try:
                        self.add_content_tab(
                            m.graph_descrip("nodes"),
                            sheetname="Nesting",
                            heading="Nesting Node List",
                        )
                    except Exception:
                        if on_error == "raise":
                            raise

        if embed:
            self.add_metadata("_self_", model)

    def add_metadata(self, key, value):
        if not isinstance(key, str):
            raise ValueError(f"metadata key must be str not {type(key)}")
        import base64

        try:
            import cloudpickle as pickle
        except ImportError:
            import pickle
        encoded_value = base64.standard_b64encode(pickle.dumps(value)).decode()
        row = self.sheet_startrow.get("_metadata_", 0)
        self.metadatasheet.write(row, 0, key)
        c = 2
        while len(encoded_value):
            self.metadatasheet.write(
                row, c, encoded_value[:30_000]
            )  # excel has cell text length cap
            c += 1
            encoded_value = encoded_value[30_000:]
        self.metadatasheet.write(row, 1, c - 2)
        self.sheet_startrow["_metadata_"] = row + 1

    def log(self, message):
        t = time.strftime("%Y-%b-%d %H:%M:%S")
        row = self.sheet_startrow.get("_log_", 0)
        self.logsheet.write(row, 0, t)
        self.logsheet.write(row, 1, str(message))
        self.sheet_startrow["_log_"] = row + 1

    def add_worksheet(self, name, force=False, hide=None):
        if not force and name in self.sheets:
            return self.sheets[name]

        s = None
        try:
            s = self.book.add_worksheet(name)
        except xlsxwriter.exceptions.DuplicateWorksheetName:
            i = 2
            while s is None:
                try:
                    s = self.book.add_worksheet(f"{name}{i}")
                except xlsxwriter.exceptions.DuplicateWorksheetName:
                    i += 1
                    if i > 99:
                        raise
        self.sheets[name] = s
        if hide is not None:
            if hide:
                s.hide()
            else:
                s.hidden = 0

        return s

    def _add_toc_entry(self, heading, target_sheet, target_row):
        self.tocsheet.write_url(
            self.sheet_startrow["Contents"],
            0,
            f"internal:'{target_sheet}'!A{target_row+1}",
            string=heading,
        )
        self.sheet_startrow["Contents"] += 1

    def add_content_tab(
        self, content, sheetname, heading=None, startrow=None, blurb=None, to_excel=None
    ):
        if startrow is None:
            startrow = self.sheet_startrow.get(sheetname, 0)
        worksheet = self.add_worksheet(sheetname)

        content_in = content
        success = False
        if to_excel is None:
            to_excel = {}

        logger.debug("writing to %s", sheetname)

        if isinstance(content, (pd.DataFrame, Styler)):
            logger.debug("writing as DataFrame (%s)", type(content))

            if isinstance(content, Styler):
                content_, content_data = content, content.data
            else:
                content_, content_data = content, content

            if heading is not None:
                worksheet.write_url(
                    startrow,
                    0,
                    "internal:Contents!A1",
                    self.toc_link_fmt,
                    string="<< Back to Table of Contents",
                )
                startrow += 1
                h = self.TAB(heading)
                worksheet.write(startrow, 0, h, self.head_fmt)
                self._add_toc_entry(h, sheetname, startrow)
                startrow += 1
                logger.debug(" after heading row is %d", startrow)
            content_.to_excel(
                self, sheet_name=worksheet.name, startrow=startrow, **to_excel
            )
            startrow += len(content_data) + content.columns.nlevels
            if content_data.index.name:
                startrow += 1
            startrow += 2  # gap
            logger.debug(" after table row is %d", startrow)
            success = True

            if sheetname not in self._col_widths:
                self._col_widths[sheetname] = {}
            if content_data.index.nlevels == 1:
                current_width = self._col_widths[sheetname].get(0, 8)
                new_width = max(
                    current_width, max(len(str(i)) for i in content_data.index)
                )
                self._col_widths[sheetname][0] = new_width
                worksheet.set_column(0, 0, new_width)
            else:
                for n in range(content_data.index.nlevels):
                    current_width = self._col_widths[sheetname].get(n, 8)
                    new_width = max(
                        current_width,
                        max(len(str(i)) for i in content_data.index.levels[n]),
                    )
                    self._col_widths[sheetname][n] = new_width
                    worksheet.set_column(n, n, new_width)

        # Render matplotlib.Figure into an Elem
        if not success and "matplotlib" in str(type(content)):
            logger.debug("writing as matplotlib (%s)", type(content))
            if isinstance(content, matplotlib.axes.Axes):
                content = content.get_figure()
            if isinstance(content, matplotlib.figure.Figure):
                try:
                    content = make_png(
                        content,
                        dpi="figure",
                        compress=True,
                        output="Elem",
                        facecolor="w",
                    )
                except Exception:
                    pass

        # Extract PNG data from Elem if found there
        if not success and isinstance(content, Elem):
            logger.debug("writing as Elem (%s)", type(content))
            if content.tag == "img":
                _v = content
            else:
                _v = content.find("img")
            if _v is not None:
                try:
                    dpi = _v.attrib["dpi"]
                except Exception:
                    dpi = None

                try:
                    _v = _v.attrib["src"]
                except Exception:
                    pass
                else:
                    if isinstance(_v, str) and _v[:22] == "data:image/png;base64,":
                        fp, img_size, dpi = make_png(
                            base64.standard_b64decode(_v[22:]),
                            dpi=dpi,
                            output="bytesio",
                            return_size=True,
                            return_dpi=True,
                        )
                        if heading is not None:
                            worksheet.write_url(
                                startrow,
                                0,
                                "internal:Contents!A1",
                                self.toc_link_fmt,
                                string="<< Back to Table of Contents",
                            )
                            startrow += 1
                            h = self.FIG(heading)
                            worksheet.write(startrow, 0, h, self.head_fmt)
                            self._add_toc_entry(h, sheetname, startrow)
                            startrow += 1
                            logger.debug(" after heading row is %d", startrow)
                        worksheet.insert_image(
                            startrow, 0, f"image-{sheetname}.png", {"image_data": fp}
                        )
                        startrow += int(np.ceil(img_size[1] / dpi[1] * 72 / 15)) + 3
                        logger.debug(" after image row is %d", startrow)
                        success = True

        if not success and isinstance(content, dict):
            logger.debug("writing as dict (%s)", type(content))
            max_row = 0
            max_col = 0
            row = -1
            prev_col = 0
            cellcontent = []
            from collections.abc import Iterable

            def unpacker(i, indent=0):
                nonlocal row, prev_col, max_row, max_col, cellcontent
                if isinstance(i, dict):
                    for k, v in i.items():
                        unpacker(k, indent)
                        unpacker(v, indent + 1)
                elif isinstance(i, pd.Series):
                    for k, v in dict(i).items():
                        unpacker(k, indent)
                        unpacker(v, indent + 1)
                elif not isinstance(i, str) and isinstance(i, Iterable):
                    for k in i:
                        unpacker(k, indent)
                else:
                    if prev_col >= indent:
                        row = row + 1
                    cellcontent.append((indent, row, i))
                    max_col = max(max_col, indent)
                    max_row = max(max_row, row)
                    prev_col = indent

            unpacker(content)
            table_cache = pd.DataFrame(
                data="",
                index=pd.RangeIndex(max_row + 1),
                columns=pd.RangeIndex(max_col + 1),
            )
            for c in cellcontent:
                table_cache.iloc[c[1], c[0]] = c[2]
            self.add_content_tab(
                table_cache,
                sheetname=sheetname,
                heading=heading,
                startrow=startrow,
                blurb=blurb,
                to_excel=dict(
                    index=False,
                    header=False,
                ),
            )
            if sheetname not in self._col_widths:
                self._col_widths[sheetname] = {}
            for n in range(max_col):
                current_width = self._col_widths[sheetname].get(n, 8)
                new_width = max(current_width, 16)
                self._col_widths[sheetname][n] = new_width
                worksheet.set_column(n, n, new_width)
            success = True

        if not success:
            import warnings

            warnings.warn(
                f"content not written to sheet {sheetname}, type {type(content_in)}",
                stacklevel=2,
            )
        else:
            if blurb is not None:
                import textwrap

                blurb = textwrap.wrap(blurb)
                for line in blurb:
                    worksheet.write(startrow, 0, line)
                    startrow += 1
                logger.debug(" after blurb row is %d", startrow)

        self.sheet_startrow[worksheet.name] = startrow
        return self._output_renderer(content_in)

    def save(self, makedirs=True, overwrite=False):
        if self.path is not None:
            if makedirs:
                import os

                dirname = os.path.dirname(self.path)
                if dirname:
                    os.makedirs(dirname, exist_ok=True)

            if not overwrite and not getattr(
                self, "__file_archived", False
            ):  # don't move twice
                from xmle.file_util import archive_existing_file

                try:
                    new_name = archive_existing_file(
                        self.path, archive_path=None, tag="creation"
                    )
                except FileNotFoundError:
                    pass
                else:
                    self.log(f"archived existing file to {new_name}")
                    setattr(self, "__file_archived", new_name)

        self.log("saving")
        super().save()


if xlsxwriter is not None:
    from pandas.io.excel import register_writer

    register_writer(ExcelWriter)


def _make_excel_writer(model, filename, save_now=True, **kwargs):
    """
    Write the model to an Excel file.

    Parameters
    ----------
    model : larch.Model or larch.ModelGroup
    filename : str
    save_now : bool, default True
        Save the model immediately.  Set to False if you want to
        write additional figures or tables to the file before saving.
    **kwargs
        Other keyword arguments are passed to the `ExcelWriter`
        constructor.

    Returns
    -------
    larch.util.excel.ExcelWriter
    """
    if xlsxwriter is None:
        raise RuntimeError("xlsxwriter is not installed")
    xl = ExcelWriter(filename, engine="xlsxwriter_larch", model=model, **kwargs)
    if save_now:
        xl.save()
    return xl


Model.to_xlsx = _make_excel_writer


def load_metadata(xlsx_filename, key=None):
    import base64
    import pickle

    import pandas

    raw = pandas.read_excel(
        xlsx_filename, sheet_name="_metadata_", index_col=0, engine="openpyxl"
    )
    metadata = {}
    if key is None:
        for key, row in raw.iterrows():
            v = row[1 : row[0] + 1].str.cat()
            metadata[key] = pickle.loads(base64.standard_b64decode(v.encode()))
        return metadata
    else:
        row = raw.loc[key]
        v = row[1 : row[0] + 1].str.cat()
        return pickle.loads(base64.standard_b64decode(v.encode()))


def _estimation_statistics_excel(
    model,
    xlsxwriter,
    sheetname,
    start_row=0,
    start_col=0,
    buffer_cols=0,
    compute_loglike_null=True,
):
    """
    Write a tabular summary of estimation statistics to excel.

    This will generate a small table of estimation statistics,
    containing:

    *	Log Likelihood at Convergence
    *	Log Likelihood at Null Parameters (if known)
    *	Log Likelihood with No Model (if known)
    *	Log Likelihood at Constants Only (if known)

    Additionally, for each included reference value (i.e.
    everything except log likelihood at convergence) the
    rho squared with respect to that value is also given.

    Each statistic is reported in aggregate, as well as
    per case.

    Parameters
    ----------
    xlsxwriter : ExcelWriter
    sheetname : str
    start_row, start_col : int
        Zero-based index of upper left cell
    buffer_cols : int
        Number of extra columns between statistic label and
        values.
    compute_loglike_null : bool, default True
        If the log likelihood at null values has not already
        been computed (i.e., if it is not cached) then compute
        it, cache its value, and include it in the output.

    """
    try:
        from ..exceptions import MissingDataError

        if start_row is None:
            start_row = xlsxwriter.sheet_startrow.get(sheetname, 0)
        worksheet = xlsxwriter.add_worksheet(sheetname)

        row = start_row

        fixed_2 = xlsxwriter.book.add_format({"num_format": "#,##0.00"})
        fixed_4 = xlsxwriter.book.add_format({"num_format": "0.0000"})
        comma_0 = xlsxwriter.book.add_format({"num_format": "#,##0"})
        bold = xlsxwriter.book.add_format({"bold": True})
        bold_centered = xlsxwriter.book.add_format({"bold": True})

        fixed_2.set_align("center")
        fixed_4.set_align("center")
        comma_0.set_align("center")
        bold_centered.set_align("center")
        bold.set_border(1)
        bold_centered.set_border(1)

        datum_col = start_col + buffer_cols

        def catname(j):
            nonlocal row, start_col, buffer_cols
            if buffer_cols:
                worksheet.merge_range(
                    row, start_col, row, start_col + buffer_cols, j, bold
                )
            else:
                worksheet.write(row, start_col, j, bold)

        catname("Statistic")
        worksheet.write(row, datum_col + 1, "Aggregate", bold_centered)
        worksheet.write(row, datum_col + 2, "Per Case", bold_centered)
        row += 1

        try:
            ncases = model.n_cases
        except MissingDataError:
            ncases = None

        catname("Number of Cases")
        if ncases:
            worksheet.merge_range(
                row, datum_col + 1, row, datum_col + 2, ncases, cell_format=comma_0
            )
        else:
            worksheet.merge_range(
                row, datum_col + 1, row, datum_col + 2, "not available"
            )
        row += 1

        mostrecent = model._most_recent_estimation_result
        if mostrecent is not None:
            catname("Log Likelihood at Convergence")
            worksheet.write(
                row, datum_col + 1, mostrecent.loglike, fixed_2
            )  # "{:.2f}".format(mostrecent.loglike)
            if ncases:
                worksheet.write(
                    row, datum_col + 2, mostrecent.loglike / ncases, fixed_4
                )  # "{:.2f}".format(mostrecent.loglike/ ncases)
            else:
                worksheet.write(row, datum_col + 2, "na")
            row += 1

        ll_z = model._cached_loglike_null
        if ll_z == 0:
            if compute_loglike_null:
                try:
                    ll_z = model.loglike_null()
                except MissingDataError:
                    pass
                else:
                    model.loglike()
            else:
                ll_z = 0
        if ll_z != 0:
            catname("Log Likelihood at Null Parameters")
            worksheet.write(row, datum_col + 1, ll_z, fixed_2)  # "{:.2f}".format(ll_z)
            if ncases:
                worksheet.write(
                    row, datum_col + 2, ll_z / ncases, fixed_4
                )  # "{:.2f}".format(ll_z/ ncases)
            else:
                worksheet.write(row, datum_col + 2, "na")
            if mostrecent is not None:
                row += 1
                catname("Rho Squared w.r.t. Null Parameters")
                rsz = 1.0 - (mostrecent.loglike / ll_z)
                worksheet.merge_range(
                    row, datum_col + 1, row, datum_col + 2, rsz, cell_format=fixed_4
                )  # "{:.3f}".format(rsz)
            row += 1

        ll_nil = model._cached_loglike_nil
        if ll_nil != 0:
            catname("Log Likelihood with No Model")
            worksheet.write(
                row, datum_col + 1, ll_nil, fixed_2
            )  # "{:.2f}".format(ll_nil)
            if ncases:
                worksheet.write(
                    row, datum_col + 2, ll_nil / ncases, fixed_4
                )  # "{:.2f}".format(ll_nil/ ncases)
            else:
                worksheet.write(row, datum_col + 2, "na")
            if mostrecent is not None:
                row += 1
                catname("Rho Squared w.r.t. No Model")
                rsz = 1.0 - (mostrecent.loglike / ll_nil)
                worksheet.merge_range(
                    row, datum_col + 1, row, datum_col + 2, rsz, cell_format=fixed_4
                )  # "{:.3f}".format(rsz)
            row += 1

        ll_c = model._cached_loglike_constants_only
        if ll_c != 0:
            catname("Log Likelihood at Constants Only")
            worksheet.write(row, datum_col + 1, ll_c, fixed_2)  # "{:.2f}".format(ll_c)
            if ncases:
                worksheet.write(
                    row, datum_col + 2, ll_c / ncases, fixed_4
                )  # "{:.2f}".format(ll_c/ ncases)
            else:
                worksheet.write(row, datum_col + 2, "na")
            if mostrecent is not None:
                row += 1
                catname("Rho Squared w.r.t. Constants Only")
                rsc = 1.0 - (mostrecent.loglike / ll_c)
                worksheet.merge_range(
                    row, datum_col + 1, row, datum_col + 2, rsc, cell_format=fixed_4
                )  # "{:.3f}".format(rsc)
            row += 1

        if mostrecent is not None:
            if "message" in mostrecent:
                catname("Optimization Message")
                worksheet.write(row, datum_col + 1, mostrecent.message)
                row += 1

        if sheetname not in xlsxwriter._col_widths:
            xlsxwriter._col_widths[sheetname] = {}
        current_width = xlsxwriter._col_widths[sheetname].get(start_col, 8)
        proposed_width = 28
        if buffer_cols:
            for b in range(buffer_cols):
                proposed_width -= xlsxwriter._col_widths[sheetname].get(
                    start_col + 1 + b, 8
                )
        new_width = max(current_width, proposed_width)
        xlsxwriter._col_widths[sheetname][start_col] = new_width
        worksheet.set_column(start_col, start_col, new_width)

        row += 2  # gap
        xlsxwriter.sheet_startrow[worksheet.name] = row
        return row
    except Exception:
        logger.exception("error in _estimation_statistics_excel")
        raise



import numpy as np
import pandas as pd
import xarray as xr

from .._optional import jax, jnp
from ..compiled import jitmethod, reset_compiled_methods
from ..dataset import Dataset, DataTree
from ..exceptions import MissingDataError
from ..folding import dissolve_zero_variance, fold_dataset
from ..optimize import OptimizeMixin
from .basemodel import MANGLE_DATA
from .basemodel import BaseModel as _BaseModel
from .jaxmodel import PanelMixin, _get_jnp_array
from .mixtures import MixtureList


class LatentClass(_BaseModel, OptimizeMixin, PanelMixin):
    compute_engine = "jax"

    def __init__(
        self, classmodel, choicemodels, datatree=None, float_dtype=np.float32, **kwargs
    ):
        self._model_subtype = "latent-class"
        classmodel._model_subtype = "class-membership"
        for k, m in choicemodels.items():
            m.ident = k
        PanelMixin.__init__(self, **kwargs)
        self._is_mangled = True
        self._dataset = None
        choicemodels_keys = sorted(choicemodels.keys())
        if (
            classmodel.datatree is None
            and isinstance(datatree, Dataset)
            and self.groupid
        ):
            df = datatree.to_dataframe()
            df["ingroup"] = df.groupby(self.groupid).cumcount() + 1
            classdata = dissolve_zero_variance(
                df.set_index([self.groupid, "ingroup"], drop=True).to_xarray(),
                "ingroup",
            )
            classdata = classdata.dc.set_altids(choicemodels_keys).drop_dims("ingroup")
            classdata.dc.CASEID = self.groupid
            classmodel.datatree = classdata
        else:
            pass
        self.ident = "latent-class"
        super().__init__(
            datatree=datatree,
            submodels=choicemodels,
            named_submodels={"classmodel": classmodel},
        )
        # self.datatree = datatree
        if classmodel.dataset is not None:
            assert sorted(classmodel.dataset.dc.altids()) == choicemodels_keys
        if float_dtype is not None:
            for v in self._models.values():
                v.float_dtype = float_dtype
        for _k, m in self._models.items():
            if m._model_subtype == "latent-class":
                pass
            elif m._model_subtype == "class-membership":
                m.groupid = self.groupid
            else:
                m.datatree = self.datatree
                m.groupid = self.groupid

    @property
    def _models(self):
        return self._parameter_bucket._models

    def save(self, filename, format="yaml", overwrite=False):
        from .saving import save_model

        return save_model(self, filename, format=format, overwrite=overwrite)

    @classmethod
    def from_dict(cls, content):
        _models = content.get("_models")
        from .saving import load_model

        classmodel = None
        choicemodels = {}
        for k, v in _models.items():
            if v._model_subtype == "class-membership":
                classmodel = load_model(v)
            else:
                choicemodels[k] = load_model(v)
        self = cls(classmodel, choicemodels)

        def loadthis(attr, wrapper=None, injector=None):
            i = content.get(attr, None)
            if i is not None:
                try:
                    if wrapper is not None:
                        i = wrapper(i)
                except AttributeError:
                    pass
                else:
                    if injector is None:
                        setattr(self, attr, i)
                    else:
                        injector(i)

        loadthis("float_dtype", lambda i: getattr(np, i))
        loadthis("compute_engine")
        loadthis("index_name")
        loadthis("parameters", xr.Dataset.from_dict, self.update_parameters)
        loadthis("availability_any")
        loadthis("availability_ca_var")
        loadthis("availability_co_vars")
        loadthis("choice_any")
        loadthis("choice_ca_var")
        loadthis("choice_co_code")
        loadthis("choice_co_vars")
        loadthis("constraint_intensity")
        loadthis("constraint_sharpness")
        loadthis("constraints")
        from .tree import NestingTree

        loadthis("graph", NestingTree.from_dict)
        loadthis("groupid")
        loadthis("logsum_parameter")
        loadthis("quantity_ca")
        loadthis("quantity_scale")
        loadthis("title")
        loadthis("utility_ca")
        loadthis("utility_co")
        loadthis("weight_co_var")
        loadthis("weight_normalization")
        return self

    @property
    def pf(self):
        return self.parameters.to_dataframe()

    @jitmethod
    def jax_probability(self, params):
        classmodel = self.class_membership_model
        pr_parts = []
        for n, k in enumerate(classmodel.dataset.dc.altids()):
            class_pr = jnp.expand_dims(classmodel.jax_probability(params)[..., n], -1)
            inclass_pr = self._models[k].jax_probability(params)
            if inclass_pr.ndim < class_pr.ndim:
                inclass_pr = inclass_pr.reshape(
                    class_pr.shape[0],
                    inclass_pr.shape[0] // class_pr.shape[0],
                    *inclass_pr.shape[1:],
                )
            print(f"{class_pr.shape=}")
            print(f"{inclass_pr.shape=}")
            pr_parts.append(inclass_pr * class_pr)
        return sum(pr_parts)

    @jitmethod
    def jax_loglike_casewise(self, params):
        n_alts = self.dataset.dc.n_alts
        ch = jnp.asarray(self.dataset["ch"])
        if ch.ndim == 2:
            pr = self.jax_probability(params)
            masked_pr = jnp.where(ch[..., :n_alts] > 0, pr[..., :n_alts], 1.0)
            log_pr = jnp.log(masked_pr)
            return (log_pr[..., :n_alts] * ch[..., :n_alts]).sum()
        elif ch.ndim >= 3:
            classmodel = self.class_membership_model
            class_pr = classmodel.jax_probability(params)
            likely_parts = []
            for n, k in enumerate(classmodel.dataset.dc.altids()):
                _yo_data = self._models[k].dataset
                k_pr = self._models[k].jax_probability(params)  # .reshape(ch.shape)
                masked_k_pr = jnp.where(ch[..., :n_alts] > 0, k_pr[..., :n_alts], 1.0)
                k_likely = jnp.power(masked_k_pr, ch[..., :n_alts]).prod([-2, -1])
                likely_parts.append(k_likely * class_pr[..., 0, n])
            return jnp.log(sum(likely_parts))

    @jitmethod
    def jax_loglike(self, params):
        return self.jax_loglike_casewise(params).sum()

    def loglike(
        self, x=None, *, start_case=None, stop_case=None, step_case=None, **kwargs
    ):
        if self.compute_engine != "jax":
            raise NotImplementedError(f"latent class with engine={self.compute_engine}")
        self.unmangle()
        if start_case is not None and start_case != 0:
            raise NotImplementedError(f"{start_case=} with engine=jax")
        if stop_case is not None and stop_case != -1:
            raise NotImplementedError(f"{stop_case=} with engine=jax")
        if step_case is not None and step_case != 1:
            raise NotImplementedError(f"{step_case=} with engine=jax")
        if x is not None:
            self.pvals = x
        result = float(self.jax_loglike(self.pvals))
        # if start_case is None and stop_case is None and step_case is None:
        #     self._check_if_best(result)
        return result

    def neg_loglike(
        self,
        x=None,
        start_case=None,
        stop_case=None,
        step_case=None,
        leave_out=-1,
        keep_only=-1,
        subsample=-1,
    ):
        result = self.loglike(
            x,
            start_case=start_case,
            stop_case=stop_case,
            step_case=step_case,
            leave_out=leave_out,
            keep_only=keep_only,
            subsample=subsample,
        )
        return -result

    @jitmethod
    def jax_d_loglike(self, params):
        return jax.grad(self.jax_loglike)(params)

    def d_loglike(
        self,
        x=None,
        *,
        start_case=None,
        stop_case=None,
        step_case=None,
        return_series=False,
        **kwargs,
    ):
        if self.compute_engine != "jax":
            raise NotImplementedError(f"latent class with engine={self.compute_engine}")
        self.unmangle()
        if start_case is not None and start_case != 0:
            raise NotImplementedError(f"{start_case=} with engine=jax")
        if stop_case is not None and stop_case != -1:
            raise NotImplementedError(f"{stop_case=} with engine=jax")
        if step_case is not None and step_case != 1:
            raise NotImplementedError(f"{step_case=} with engine=jax")
        if x is not None:
            self.pvals = x
        result = self.jax_d_loglike(self.pvals)

        # print("converge?=", jnp.max(jnp.absolute(result)))

        if return_series:
            result = pd.Series(result, index=self.pnames)
        return result

    def neg_d_loglike(self, x=None, start_case=0, stop_case=-1, step_case=1, **kwargs):
        result = self.d_loglike(
            x, start_case=start_case, stop_case=stop_case, step_case=step_case, **kwargs
        )
        return -np.asarray(result)

    def maximize_loglike(
        self,
        *args,
        **kwargs,
    ):
        """
        Maximize the log likelihood.

        Parameters
        ----------
        method : str, optional
            The optimization method to use.  See scipy.optimize for
            most possibilities, or use 'BHHH'. Defaults to SLSQP if
            there are any constraints or finite parameter bounds,
            otherwise defaults to BHHH.
        quiet : bool, default False
            Whether to suppress the dashboard.
        options : dict, optional
            These options are passed through to the `scipy.optimize.minimize`
            function.
        maxiter : int, optional
            Maximum number of iterations.  This argument is just added to
            `options` for most methods.

        Returns
        -------
        dictx
            A dictionary of results, including final log likelihood,
            elapsed time, and other statistics.  The exact items
            included in output will vary by estimation method.

        """
        from .optimization import maximize_loglike

        return maximize_loglike(self, *args, **kwargs)

    def reflow_data_arrays(self):
        """Reload the internal data_arrays so they are consistent with the datatree."""
        datatree = self.datatree
        if datatree is None:
            raise ValueError("missing datatree")

        classmodel = self.class_membership_model
        request = classmodel.required_data()
        request.pop("avail_any", None)
        for _kname, kmodel in self._models.items():
            if kmodel._model_subtype in ("class-membership", "latent-class"):
                continue
            kreq = kmodel.required_data()
            for k, v in kreq.items():
                if k not in request:
                    request[k] = v
                else:
                    if isinstance(request[k], dict):
                        request[k].update(v)
                    elif isinstance(request[k], (list | tuple)):
                        request[k] = list(set(request[k]) | set(v))
                    else:
                        if request[k] != v:
                            raise ValueError("incompatible requests")

        if isinstance(self.groupid, str):
            request["group_co"] = self.groupid

        from .data_arrays import prepare_data

        dataset, self.dataflows = prepare_data(
            datasource=datatree,
            request=request,
            float_dtype=np.float32,
            cache_dir=datatree.cache_dir,
            flows=getattr(self, "dataflows", None),
        )
        if isinstance(self.groupid, str):
            dataset = fold_dataset(dataset, "group")
        elif self.groupid is not None:
            dataset = fold_dataset(dataset, self.groupid)
        self.dataset = dataset

        for _kname, kmodel in self._models.items():
            if kmodel._model_subtype in ("class-membership", "latent-class"):
                continue
            kmodel.dataset = self.dataset
            # kmodel.reflow_data_arrays() # not full reflow, just...
            kmodel._data_arrays = kmodel.dataset.dc.to_arrays(
                kmodel.graph,
                float_dtype=kmodel.float_dtype,
            )
            # if kmodel.work_arrays is not None:
            kmodel._rebuild_fixed_arrays()
            kmodel._rebuild_work_arrays()
        classmodel_ids = [
            kid
            for (kid, km) in self._models.items()
            if km._model_subtype not in ("latent-class", "class-membership")
        ]
        classmodel_data = dissolve_zero_variance(
            self.dataset.drop_dims(self.dataset.dc.ALTID).dc.set_altids(classmodel_ids),
            "ingroup",
        )
        classmodel.dataset = classmodel_data
        classmodel.reflow_data_arrays()
        # classmodel._data_arrays = classmodel.dataset.dc.to_arrays(
        #     classmodel.graph,
        #     float_dtype=classmodel.float_dtype,
        # )
        # if classmodel.work_arrays is not None:
        #     classmodel._rebuild_work_arrays()

    def mangle(self, data=True, structure=True):
        super().mangle(data=data, structure=structure)
        reset_compiled_methods(self)
        self._is_mangled = True

    def unmangle(self, force=False, structure_only=False):
        if self._is_mangled:
            super().unmangle(force=force, structure_only=structure_only)
            if not structure_only:
                self.reflow_data_arrays()
                self._is_mangled = False
            else:
                self._is_mangled = MANGLE_DATA

    @property
    def dataset(self):
        """larch.Dataset : Data arrays as loaded for model computation."""
        super().unmangle()
        if self._dataset is None:
            self.reflow_data_arrays()
        try:
            return self._dataset
        except AttributeError:
            return None

    @dataset.setter
    def dataset(self, dataset):
        if dataset is self._dataset:
            return
        from xarray import Dataset as _Dataset

        if isinstance(dataset, Dataset):
            self._dataset = dataset
            self._data_arrays = None
            self.mangle()
        elif isinstance(dataset, _Dataset):
            self._dataset = Dataset(dataset)
            self._data_arrays = None
            self.mangle()
        else:
            raise TypeError(f"dataset must be Dataset not {type(dataset)}")

    @property
    def datatree(self):
        """DataTree : A source for data for the model."""
        try:
            return self._datatree
        except AttributeError:
            return None

    @datatree.setter
    def datatree(self, tree):
        if tree is self.datatree:
            return
        if isinstance(tree, DataTree) or tree is None:
            self._datatree = tree
            self.mangle()
        elif isinstance(tree, Dataset):
            self._datatree = tree.dc.as_tree()
            self.mangle()
        else:
            try:
                self._datatree = DataTree(main=Dataset.construct(tree))
            except Exception as err:
                raise TypeError(f"datatree must be DataTree not {type(tree)}") from err
            else:
                self.mangle()

    @property
    def data_as_loaded(self):
        return self._dataset

    @property
    def class_membership_model(self):
        # TODO optimize this
        for m in self._models.values():
            if m._model_subtype == "class-membership":
                return m

    def total_weight(self):
        """
        Compute the total weight of cases in the loaded data.

        Returns
        -------
        float
        """
        if self.class_membership_model._data_arrays is not None:
            return self.class_membership_model._data_arrays.wt.sum()
        raise MissingDataError("no data_arrays are set")

    def logloss(
        self,
        x=None,
        start_case=None,
        stop_case=None,
        step_case=None,
        leave_out=-1,
        keep_only=-1,
        subsample=-1,
    ):
        result = self.loglike(
            x,
            start_case=start_case,
            stop_case=stop_case,
            step_case=step_case,
            leave_out=leave_out,
            keep_only=keep_only,
            subsample=subsample,
        )
        return -result / self.total_weight()

    def d_logloss(self, x=None, start_case=0, stop_case=-1, step_case=1, **kwargs):
        result = self.d_loglike(
            x, start_case=start_case, stop_case=stop_case, step_case=step_case, **kwargs
        )
        return -np.asarray(result) / self.total_weight()

    def simple_fit_bhhh(self, *args, **kwargs):
        raise NotImplementedError()

    def calculate_parameter_covariance(self, pvals=None, *, robust=False):
        if pvals is None:
            pvals = self.pvals
        locks = np.asarray(self.pholdfast.astype(bool))
        if self.compute_engine == "jax":
            se, hess, ihess = self.jax_param_cov(pvals)
        else:
            raise NotImplementedError(f"compute_engine={self.compute_engine}")
            # hess = -self.d2_loglike(pvals)
            # if self.parameters["holdfast"].sum():
            #     free = self.pholdfast == 0
            #     hess_ = hess[free][:, free]
            #     ihess_ = np.linalg.inv(hess_)
            #     ihess = _arr_inflate(ihess_, locks)
            # else:
            #     ihess = np.linalg.inv(hess)
            # se = np.sqrt(ihess.diagonal())
            # self.pstderr = se
        hess = np.asarray(hess).copy()
        hess[locks, :] = 0
        hess[:, locks] = 0
        ihess = np.asarray(ihess).copy()
        ihess[locks, :] = 0
        ihess[:, locks] = 0
        self.add_parameter_array("hess", hess)
        self.add_parameter_array("ihess", ihess)

        # constrained covariance
        if self.constraints:
            constraints = list(self.constraints)
        else:
            constraints = []
        try:
            constraints.extend(self._get_bounds_constraints())
        except AttributeError:
            pass

        if constraints:
            binding_constraints = list()
            self.add_parameter_array("unconstrained_std_err", self.pstderr)
            self.add_parameter_array("unconstrained_covariance_matrix", ihess)

            s = np.asarray(ihess)
            pvals = self.pvals
            for c in constraints:
                if np.absolute(c.fun(pvals)) < c.binding_tol:
                    binding_constraints.append(c)
                    b = c.jac(self.pf.value)
                    den = b @ s @ b
                    if den != 0:
                        s = s - (1 / den) * s @ b.reshape(-1, 1) @ b.reshape(1, -1) @ s
            self.add_parameter_array("covariance_matrix", s)
            self.pstderr = np.sqrt(s.diagonal())

            # Fix numerical issues on some constraints, add constrained notes
            if binding_constraints or any(self.pholdfast != 0):
                notes = {}
                for c in binding_constraints:
                    pa = c.get_parameters()
                    for p in pa:
                        # if self.pf.loc[p, 't_stat'] > 1e5:
                        #     self.pf.loc[p, 't_stat'] = np.inf
                        #     self.pf.loc[p, 'std_err'] = np.nan
                        # if self.pf.loc[p, 't_stat'] < -1e5:
                        #     self.pf.loc[p, 't_stat'] = -np.inf
                        #     self.pf.loc[p, 'std_err'] = np.nan
                        n = notes.get(p, [])
                        n.append(c.get_binding_note(pvals))
                        notes[p] = n
                constrained_note = (
                    pd.Series({k: "\n".join(v) for k, v in notes.items()}, dtype=object)
                    .reindex(self.pnames)
                    .fillna("")
                )
                constrained_note[self.pholdfast != 0] = "fixed value"
                self.add_parameter_array("constrained", constrained_note)

        if robust:
            self.robust_covariance()
            se = self.parameters["robust_std_err"]

        return se, hess, ihess

    def robust_covariance(self):
        raise NotImplementedError()


class MixedLatentClass(LatentClass):
    mixtures = MixtureList()

    def __init__(
        self, *args, n_draws=100, prerolled_draws=True, common_draws=False, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self._n_draws = n_draws
        self._draws = None
        self.prerolled_draws = prerolled_draws
        self.common_draws = common_draws

    @classmethod
    def from_dict(cls, content):
        self = super().from_dict(content)

        def loadthis(attr, wrapper=None, injector=None):
            i = content.get(attr, None)
            if i is not None:
                try:
                    if wrapper is not None:
                        i = wrapper(i)
                except AttributeError:
                    pass
                else:
                    if injector is None:
                        setattr(self, attr, i)
                    else:
                        injector(i)

        loadthis("mixtures", self.mixtures.from_list)
        loadthis("n_draws")
        loadthis("prerolled_draws")
        loadthis("common_draws")

    def apply_random_draws(self, parameters, draws=None):
        # if draws is None:
        #     draws = self._draws
        parameters = jnp.broadcast_to(
            parameters, [*draws.shape[:-1], parameters.shape[0]]
        )
        for mix_n, mix in enumerate(self.mixtures):
            u = draws[..., mix_n]
            parameters = mix.roll(u, parameters)
        return parameters

    def _make_random_draws_out(self, n_draws, n_mixtures, random_key):
        draws = jax.random.uniform(
            random_key,
            [n_draws, n_mixtures],
        )

        def body(i, carry):
            draws, rnd_key = carry
            rnd_key, subkey = jax.random.split(rnd_key)
            draws = draws.at[:, i].add(jax.random.permutation(subkey, draws.shape[0]))
            return draws, rnd_key

        draws, random_key = jax.lax.fori_loop(
            0, draws.shape[1], body, (draws, random_key)
        )
        return draws * (1 / n_draws), random_key

    @jitmethod
    def _jax_probability_by_class(self, params, databundle):
        # classmodel = self['classmodel']
        pr_parts = []
        for _n, k in enumerate(self.class_membership_model.dataset.dc.altids()):
            pr_parts.append(
                self._models[k]._jax_probability(params, databundle)
                # * jnp.expand_dims(classmodel._jax_probability(params, databundle)[..., n], -1)
            )
        return jnp.stack(pr_parts)

    @jitmethod
    def _jax_likelihood_by_class(self, params, databundle):
        n_alts = self.dataset.dc.n_alts
        ch = databundle.get("ch", None)[..., :n_alts]
        pr_parts = self._jax_probability_by_class(params, databundle)[..., :n_alts]
        likely = jnp.where(
            jnp.expand_dims(ch, 0), pr_parts, 1.0
        )  # TODO make power if needed
        return likely

    @jitmethod(static_argnums=(3,), static_argnames="n_draws")
    def _jax_loglike_casewise_mixed(
        self, params, databundle, groupbundle=None, n_draws=100
    ):
        classmodel = self.class_membership_model
        if self.prerolled_draws:
            draws = groupbundle.get("draws", None)
        else:
            rk = groupbundle.get("rk", None)
            draws, _ = self._make_random_draws_out(n_draws, len(self.mixtures), rk)
        ch = databundle.get("ch", None)
        rand_params = self.apply_random_draws(params, draws)
        if ch.ndim == 2:  # PANEL DATA
            # vmap over ingroup
            likelihood_f = jax.vmap(
                self._jax_likelihood_by_class,
                in_axes=(None, 0),
            )
            # vmap over draws
            likelihood = jax.vmap(
                likelihood_f,
                in_axes=(0, None),
                out_axes=-1,
            )(rand_params, databundle)
            # collapse likelihood over all alternatives over all cases-within-panel
            likelihood = likelihood.prod([0, 2])
            # likelihood.shape is now (n_classes, n_draws)
            ################
            class_pr = jax.vmap(
                classmodel._jax_probability,
                in_axes=(0, None),
                out_axes=-1,
            )(
                rand_params,
                {
                    "co": databundle["co"][0],
                },
            )
            # class_pr.shape = (nclasses, ndraws)
            meta_likely = (likelihood * class_pr).sum(0).mean(0)
            return jnp.log(meta_likely)
        else:
            # vmap over draws
            likelihood = jax.vmap(
                self._jax_likelihood_by_class,
                in_axes=(0, None),
                out_axes=-1,
            )(rand_params, databundle)
            # collapse likelihood over all alternatives
            likelihood = likelihood.prod(0)
            # average over all draws
            likelihood = likelihood.mean(0)
            return jnp.log(likelihood)

    @jitmethod
    def jax_loglike_casewise(self, params):
        if len(self.mixtures) == 0:
            return super().jax_loglike_casewise(params)
        ca = _get_jnp_array(self.dataset, "ca")
        co = _get_jnp_array(self.dataset, "co")
        av = _get_jnp_array(self.dataset, "av")
        ch = _get_jnp_array(self.dataset, "ch")
        n_draws = self._n_draws
        seed = getattr(self, "seed", 42)
        if av is not None:
            depth = av.ndim - 1
            shape = av.shape[:-1]
        elif co is not None:
            depth = co.ndim - 1
            shape = co.shape[:-1]
        elif ca is not None:
            depth = ca.ndim - 2
            shape = ca.shape[:-2]
        elif ch is not None:
            depth = ch.ndim - 1
            shape = ch.shape[:-1]
        else:
            raise ValueError("missing data")
        random_key = jax.random.PRNGKey(seed)
        if self.groupid is not None:
            depth = depth - 1
            shape = shape[:-1]
        f = (
            self._jax_loglike_casewise_mixed
        )  # params, databundle, groupbundle=None, n_draws=100
        from .random import keysplit

        commons = None if self.common_draws else 0
        for _i in range(depth):
            f = jax.vmap(f, in_axes=(None, 0, commons, None))
            if not self.prerolled_draws:
                random_key, shape = keysplit(random_key, shape)
        if self.prerolled_draws:
            return f(
                params,
                {"ca": ca, "co": co, "av": av, "ch": ch},
                {"draws": self._draws},
                n_draws,
            )
        else:
            return f(
                params,
                {"ca": ca, "co": co, "av": av, "ch": ch},
                {"rk": random_key},
                n_draws,
            )

    def make_random_draws(self, engine="numpy"):
        self.unmangle()
        for i in self.mixtures:
            i.prep(self)
        n_panels = self.dataset.dc.n_panels
        n_mixtures = len(self.mixtures)
        n_draws = self._n_draws
        draws = None
        if self._draws is not None:
            if self.common_draws and self._draws.shape == (n_draws, n_mixtures):
                draws = self._draws
            if not self.common_draws and self._draws.shape == (
                n_panels,
                n_draws,
                n_mixtures,
            ):
                draws = self._draws
        if draws is None:
            if engine == "numpy":
                seed = getattr(self, "seed", 0)
                if self.common_draws:
                    if n_draws > 0 and n_mixtures > 0:
                        draws, seed = self._make_random_draws_numpy(
                            n_draws, n_mixtures, seed
                        )
                else:
                    if n_draws > 0 and n_mixtures > 0 and n_panels > 0:
                        draws, seed = self._make_random_draws_numpy_2(
                            n_draws, n_mixtures, n_panels, seed
                        )
                    else:
                        draws = None
            elif engine == "jax":
                seed = getattr(self, "seed", 0)
                rk = jax.random.PRNGKey(seed)
                if self.common_draws:
                    if n_draws > 0 and n_mixtures > 0:
                        draws = self._make_random_draws_out(n_draws, n_mixtures, rk)[0]
                else:
                    if n_draws > 0 and n_mixtures > 0 and n_panels > 0:
                        draws = self._make_random_draws_out_2(
                            n_draws, n_mixtures, n_panels, rk
                        )[0]
                    else:
                        draws = None
            else:
                raise ValueError(f"unknown random engine {engine!r}")
            if self.prerolled_draws:
                self._draws = draws
        return draws

    # @jitmethod(static_argnums=(0,1,2), static_argnames=('n_draws', 'n_mixtures', 'n_cases'))
    def _make_random_draws_out_2(self, n_draws, n_mixtures, n_cases, random_key):
        def body(carry, x):
            rkey = carry
            draws, rkey = self._make_random_draws_out(n_draws, n_mixtures, rkey)
            return rkey, draws

        random_key, draws = jax.lax.scan(body, random_key, None, n_cases)
        return draws, random_key



import time

import numpy as np
from IPython.display import HTML, display
from rich.live import Live
from rich.table import Table


class Dashboard:
    def __init__(self, status="", params=None, show=True, throttle=0.1):
        self.title = "[bold blue]Larch Model Dashboard"
        self.status = status
        self.loglike = "Log Likelihood = [red italic]pending"
        if params is None:
            params = "[red]No parameters"
        self.params = params
        self._live = None
        self.display(show)
        self.tick = time.time()
        self._throttle = throttle

    def display(self, show=True):
        if show:
            self.tag = display(self.render(), display_id=True)
        else:
            self.clear()
            self.tag = None

    def set_title(self, x):
        if x is not None:
            self.title = f"[bold blue]{x}"

    def set_loglike(self, x, best=None):
        if x is not None:
            if not np.isfinite(x):
                if best is not None:
                    self.loglike = f"Log Likelihood [red]Current = {x:17.6f}"
                else:
                    self.loglike = f"Log Likelihood = [red]{x:24.6f}"
            else:
                if best is not None:
                    self.loglike = f"Log Likelihood [yellow]Current = {x:17.6f}"
                else:
                    self.loglike = f"Log Likelihood = {x:24.6f}"
        if best is not None:
            self.loglike += f" [green]Best = {best:17.6f}"

    def render(self):
        t = Table()
        t.add_column("[bold blue]Larch Model Dashboard")
        if self.status:
            t.add_row(self.status)
        t.add_row(self.loglike)
        t.add_row(self.params)
        if self._live is not None:
            self._live.update(t)
        return t

    def __enter__(self):
        self._live = Live(self.render())
        self._live.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._live.__exit__(exc_type, exc_val, exc_tb)

    def update_content(
        self, title=None, loglike=None, params=None, status=None, bestloglike=None
    ):
        self.set_title(title)
        self.set_loglike(loglike, bestloglike)
        if params is not None:
            self.params = params
        if status is not None:
            self.status = status
        if self.tag is not None:
            self.tag.update(self.render())

    def clear(self):
        try:
            tag = self.tag
        except AttributeError:
            tag = None
        if tag is not None:
            self.tag.update(HTML(""))

    def throttle(self):
        now = time.time()
        if now < self.tick + self._throttle:
            return False
        else:
            self.tick = now
            return True



import numba as nb
import numpy as np
import pandas as pd


@nb.njit
def _fast_map(fm, target, dtype=np.int32):
    out = np.zeros(len(target), dtype=dtype)
    for n in range(target.size):
        out[n] = fm[target[n]]
    return out


class FastMapping:
    def __init__(self, source, to_range=np.int64):
        if isinstance(source, pd.Series):
            m = nb.typed.Dict.empty(
                key_type=nb.from_dtype(source.index.dtype),
                value_type=nb.from_dtype(source.dtype),
            )
            for k, v in source.items():
                m[k] = v
            self._in_dtype = source.index.dtype
            self._out_dtype = source.dtype
            self._mapper = m
        elif to_range:
            m = nb.typed.Dict.empty(
                key_type=nb.from_dtype(source.dtype),
                value_type=nb.from_dtype(to_range),
            )
            for v, k in enumerate(source):
                m[k] = v
            self._in_dtype = source.dtype
            self._out_dtype = to_range
            self._mapper = m
        else:
            raise ValueError("invalid input")

    def __len__(self):
        return len(self._mapper)

    def __contains__(self, item):
        return item in self._mapper

    def __getitem__(self, item):
        return self._mapper[item]

    def apply_to(self, target):
        if isinstance(target, pd.Series):
            return pd.Series(
                _fast_map(
                    self._mapper,
                    target.astype(self._in_dtype).to_numpy(),
                    dtype=self._out_dtype,
                ),
                index=target.index,
            )
        return _fast_map(
            self._mapper,
            np.asarray(target, dtype=self._in_dtype),
            dtype=self._out_dtype,
        )



import warnings

import pandas as pd
import xarray as xr
from sharrow.accessors import register_dataset_method


class NameConflictWarning(Warning):
    """Warning for conflicts in name registration."""


def _register_classmethod(name, cls):
    def decorator(classmeth):
        if hasattr(cls, name):
            warnings.warn(
                f"registration of classmethod {classmeth!r} under name {name!r} "
                f"for type {cls!r} is overriding a preexisting attribute with "
                f"the same name.",
                NameConflictWarning,
                stacklevel=2,
            )
        setattr(cls, name, classmethod(classmeth))
        return classmeth

    return decorator


def register_dataarray_classmethod(func):
    """
    Register a custom classmethod on xarray.DataArray objects.

    Use this as a decorator to add class methods.

    Parameters
    ----------
    func : Callable
        Class method to add.  The name is inferred from the
        original name of this function.
    """
    return _register_classmethod(func.__name__, xr.DataArray)(func)


def register_dataset_classmethod(func):
    """
    Register a custom classmethod on xarray.Dataset objects.

    Use this as a decorator to add class methods.

    Parameters
    ----------
    func : Callable
        Class method to add.  The name is inferred from the
        original name of this function.
    """
    return _register_classmethod(func.__name__, xr.Dataset)(func)


@register_dataset_method
def set_dtypes(self, dtypes, inplace=False, on_error="warn"):
    """
    Set the dtypes for the variables in this Dataset.

    Parameters
    ----------
    dtypes : Mapping or DataFrame
        Mapping of names to dtypes, or a DataFrame to infer such a
        mapping.
    inplace : bool, default False
        Whether to convert dtypes inplace.
    on_error : {'warn', 'raise', 'ignore'}
        What to do when a type conversion triggers an error.

    Returns
    -------
    Dataset
    """
    if isinstance(dtypes, pd.DataFrame):
        dtypes = dtypes.dtypes
    if inplace:
        obj = self
    else:
        obj = self.copy()
    for k in obj:
        if k not in dtypes:
            continue
        try:
            obj[k] = obj[k].astype(dtypes[k])
        except Exception as err:
            if on_error == "warn":
                warnings.warn(f"{err!r} on converting {k}", stacklevel=2)
            elif on_error == "raise":
                raise
    return obj



import os
from pathlib import Path

import numpy as np
import pandas as pd
import yaml

from ... import DataFrames, Model, P, X
from .. import Dict
from .general import (
    apply_coefficients,
    cv_to_ca,
    linear_utility_from_spec,
)


def size_coefficients_from_spec(size_spec):
    size_coef = size_spec.stack().reset_index()
    size_coef.index = size_coef.iloc[:, 0] + "_" + size_coef.iloc[:, 1]
    size_coef = size_coef.loc[size_coef.iloc[:, 2] > 0]
    size_coef["constrain"] = "F"
    one_each = size_coef.groupby("segment").first().reset_index()
    size_coef.loc[one_each.iloc[:, 0] + "_" + one_each.iloc[:, 1], "constrain"] = "T"
    size_coef = size_coef.iloc[:, 2:]
    size_coef.columns = ["value", "constrain"]
    size_coef.index.name = "coefficient_name"
    size_coef["value"] = np.log(size_coef["value"])
    return size_coef


def location_choice_model(
    model_selector="workplace",
    edb_directory="output/estimation_data_bundle/{model_selector}_location/",
    coefficients_file="{model_selector}_location_coefficients.csv",
    spec_file="{model_selector}_location_SPEC.csv",
    size_spec_file="{model_selector}_location_size_terms.csv",
    alt_values_file="{model_selector}_location_alternatives_combined.csv",
    chooser_file="{model_selector}_location_choosers_combined.csv",
    settings_file="{model_selector}_location_model_settings.yaml",
    landuse_file="{model_selector}_location_landuse.csv",
    return_data=False,
):
    edb_directory = edb_directory.format(model_selector=model_selector)

    def _read_csv(filename, **kwargs):
        filename = filename.format(model_selector=model_selector)
        return pd.read_csv(os.path.join(edb_directory, filename), **kwargs)

    coefficients = _read_csv(
        coefficients_file,
        index_col="coefficient_name",
    )
    spec = _read_csv(spec_file)
    alt_values = _read_csv(alt_values_file)
    chooser_data = _read_csv(chooser_file)
    landuse = _read_csv(landuse_file, index_col="zone_id")
    size_spec = _read_csv(size_spec_file)

    settings_file = settings_file.format(model_selector=model_selector)
    with open(os.path.join(edb_directory, settings_file)) as yf:
        settings = yaml.load(
            yf,
            Loader=yaml.SafeLoader,
        )
    CHOOSER_SEGMENT_COLUMN_NAME = settings["CHOOSER_SEGMENT_COLUMN_NAME"]
    SEGMENT_IDS = settings["SEGMENT_IDS"]

    # filter size spec for this location choice only
    size_spec = (
        size_spec.query(f"model_selector == '{model_selector}'")
        .drop(columns="model_selector")
        .set_index("segment")
    )
    size_spec = size_spec.loc[:, size_spec.max() > 0]

    size_coef = size_coefficients_from_spec(size_spec)

    # Remove shadow pricing and pre-existing size expression for re-estimation
    spec = (
        spec.set_index("Label")
        .drop(
            index=[
                "util_size_variable",  # pre-computed size (will be re-estimated)
                "util_utility_adjustment",  # shadow pricing (ignored in estimation)
            ]
        )
        .reset_index()
    )

    m = Model()
    if len(spec.columns) == 4:  # ['Label', 'Description', 'Expression', 'coefficient']
        m.utility_ca = linear_utility_from_spec(
            spec,
            x_col="Label",
            p_col="coefficient",
            ignore_x=("local_dist",),
        )
    else:
        m.utility_ca = linear_utility_from_spec(
            spec,
            x_col="Label",
            p_col=SEGMENT_IDS,
            ignore_x=("local_dist",),
            segment_id=CHOOSER_SEGMENT_COLUMN_NAME,
        )

    m.quantity_ca = sum(
        P(f"{i}_{q}") * X(q) * X(f"{CHOOSER_SEGMENT_COLUMN_NAME}=={SEGMENT_IDS[i]}")
        for i in size_spec.index
        for q in size_spec.columns
        if size_spec.loc[i, q] != 0
    )

    apply_coefficients(coefficients, m)
    apply_coefficients(size_coef, m, minimum=-6, maximum=6)

    x_co = chooser_data.set_index("person_id")
    x_ca = cv_to_ca(alt_values.set_index(["person_id", "variable"]))

    # label segments with names
    SEGMENT_IDS_REVERSE = {v: k for k, v in SEGMENT_IDS.items()}
    x_co["_segment_label"] = x_co[CHOOSER_SEGMENT_COLUMN_NAME].apply(
        lambda x: SEGMENT_IDS_REVERSE[x]
    )

    # compute total size values by segment
    for segment in size_spec.index:
        total_size_segment = pd.Series(0, index=landuse.index)
        x_co["total_size_" + segment] = 0
        for land_use_field in size_spec.loc[segment].index:
            total_size_segment += (
                landuse[land_use_field] * size_spec.loc[segment, land_use_field]
            )
        x_co["total_size_" + segment] = total_size_segment.loc[
            x_co["override_choice"]
        ].to_numpy()

    # for each chooser, collate the appropriate total size value
    x_co["total_size_segment"] = 0
    for segment in size_spec.index:
        labels = "total_size_" + segment
        rows = x_co["_segment_label"] == segment
        x_co.loc[rows, "total_size_segment"] = x_co[labels][rows]

    # Remove choosers with invalid observed choice (appropriate total size value = 0)
    valid_observed_zone = x_co["total_size_segment"] > 0
    x_co = x_co[valid_observed_zone]
    x_ca = x_ca[x_ca.index.get_level_values("person_id").isin(x_co.index)]

    # Merge land use characteristics into CA data
    x_ca_1 = pd.merge(x_ca, landuse, on="zone_id", how="left")
    x_ca_1.index = x_ca.index

    # Availability of choice zones
    av = x_ca_1["util_no_attractions"].apply(lambda x: False if x == 1 else True)

    d = DataFrames(
        co=x_co,
        ca=x_ca_1,
        av=av,
    )
    m.dataservice = d
    m.choice_co_code = "override_choice"

    if return_data:
        return m, Dict(
            edb_directory=Path(edb_directory),
            alt_values=alt_values,
            chooser_data=chooser_data,
            coefficients=coefficients,
            landuse=landuse,
            spec=spec,
            size_spec=size_spec,
        )

    return m



import numpy as np
import pandas as pd
import xarray as xr

from .dataset import DataArray, Dataset
from .util.bounded_kde import BoundedKDE


def fair_range(x, buffering=0.1):
    lo, hi = np.percentile(x, [1, 99])
    buffer = (hi - lo) * buffering
    return lo - buffer, hi + buffer


def scotts_factor(self, n, d=1):
    """
    Compute Scott's factor.

    Parameters
    ----------
    n : numeric
        Effective number of data points, either the cardinality
        of the array or the total weight.
    d : int, default 1
        Number of dimensions.

    Returns
    -------
    float
    """
    return np.power(n, -1.0 / (d + 4))


def silverman_factor(self, n, d=1):
    """
    Compute the Silverman factor.

    Parameters
    ----------
    n : numeric
        Effective number of data points, either the cardinality
        of the array or the total weight.
    d : int, default 1
        Number of dimensions.

    Returns
    -------
    float
    """
    return np.power(n * (d + 2.0) / 4.0, -1.0 / (d + 4))


_GENERIC = "< All Alternatives >"


def idca_statistics_data(
    var,
    bw_method="scott",
    choice=None,
    avail=None,
    wgt=None,
    scale_ch=True,
    bins=10,
    resolution=10,
):
    name = var.name or "Variable"
    if avail is None:
        avail = DataArray(xr.ones_like(var, dtype=np.int8))
    if wgt is None:
        wgt = np.ones(var.n_cases, dtype=np.float32)
    else:
        wgt = np.asarray(wgt)
    n_alts = var.n_alts
    # n = np.prod(var.shape)
    # if bw_method == "scott":
    #     bandwidth = scotts_factor(n, 1)
    # elif bw_method == "silverman":
    #     bandwidth = scotts_factor(n, 1)
    # elif np.isscalar(bw_method) and not isinstance(bw_method, str):
    #     bandwidth = bw_method
    # else:
    #     raise ValueError("`bw_method` should be 'scott', 'silverman', or a scalar")
    stats = Dataset(
        {
            "mean": var.mean(var.CASEID),
            "std": var.std(var.CASEID),
        }
    )
    data = {}
    data_ch = {}
    data_av = {}
    stats_std = {}
    for k in range(n_alts):
        if "alt_names" in var.coords:
            key = str(var.coords["alt_names"][k].data)
        elif var.ALTID in var.coords:
            key = str(var.coords[var.ALTID][k].data)
        else:
            key = str(k)
        stats_std[key] = stats["std"][k]
        data[key] = var.isel({var.ALTID: k}).values
        if choice is not None:
            data_ch[key] = choice.isel({choice.ALTID: k}).values
        data_av[key] = avail.isel({avail.ALTID: k}).values
    hist_overall, bins = np.histogram(
        var.data, bins=bins, weights=avail.data * wgt[:, None]
    )
    kde_overall = BoundedKDE(var.data, weights=avail.data * wgt[:, np.newaxis])
    # eff_wgts = avail.data * wgt[:, np.newaxis]
    # bandwidth = (
    #         weighted_sample_std(var.data.reshape(-1), eff_wgts.reshape(-1), ddof=1.0)
    #         * np.power(eff_wgts.sum(), -1.0 / 5)
    # )
    x = np.linspace(bins[0], bins[-1], resolution * len(bins))
    hist = {}
    hist_ch = {}
    kde = {}
    kde_ch = {}
    for k, v in data.items():
        hist[k] = np.append(
            np.repeat(
                np.histogram(v, bins=bins, weights=data_av[k] * wgt)[0], resolution
            ),
            0,
        )
        if stats_std[k] > 0:
            kde[k] = BoundedKDE(v, bw_method=kde_overall, weights=data_av[k] * wgt)(x)
    kde[_GENERIC] = kde_overall(x)

    if choice is not None:
        for k, v in data.items():
            hist_ch[k] = np.append(
                np.repeat(
                    np.histogram(v, bins=bins, weights=data_ch[k] * data_av[k] * wgt)[
                        0
                    ],
                    resolution,
                ),
                0,
            )
            if stats_std[k] > 0:
                y = BoundedKDE(
                    v, bw_method=kde_overall, weights=data_ch[k] * data_av[k] * wgt
                )(x)
                if scale_ch:
                    y = y * (data_ch[k] * wgt).sum() / (data_av[k] * wgt).sum()
                kde_ch[k] = y
        y = BoundedKDE(
            var.values.astype(np.float32),
            bw_method=kde_overall,
            weights=(choice.values * avail.values * wgt[:, np.newaxis]),
        )(x)
        if scale_ch:
            y = (
                y
                * (choice.values * wgt[:, np.newaxis]).sum()
                / (avail.values * wgt[:, np.newaxis]).sum()
            )
        kde_ch[_GENERIC] = y
        plot_data = pd.concat(
            [
                pd.DataFrame(hist),
                pd.DataFrame(hist_ch),
                pd.DataFrame(kde),
                pd.DataFrame(kde_ch),
            ],
            axis=1,
            keys=["Frequency", "FrequencyChosen", "Density", "DensityChosen"],
        )
        plot_data = (
            plot_data.set_index(x)
            .stack()
            .rename_axis(index=(name, "Alternative"))
            .reset_index()
        )
    else:
        kde_ch = None
        plot_data = (
            pd.DataFrame(kde, index=x)
            .stack()
            .rename("Density")
            .rename_axis(index=(name, "Alternative"))
            .reset_index()
        )
    return stats, plot_data


def idca_statistics(
    var,
    bw_method="scott",
    choice=None,
    avail=None,
    wgt=None,
    scale_ch=True,
    bins=10,
    resolution=100,
):
    import altair as alt

    stats, plot_data = idca_statistics_data(
        var,
        bw_method=bw_method,
        choice=choice,
        avail=avail,
        wgt=wgt,
        scale_ch=scale_ch,
        bins=bins,
        resolution=resolution,
    )
    name = var.name or "Variable"
    selection = alt.selection_single(fields=["Alternative"], bind="legend", empty="all")
    selection2 = alt.selection_single(
        fields=["Alternative"], bind="legend", empty="none"
    )
    ax = alt.Chart(plot_data)
    alt_names = list(str(i.data) for i in var["alt_names"])
    # A dropdown filter
    # options = [None] + list(str(i.data) for i in var['alt_names'])
    # options_dropdown = alt.binding_select(options=options)
    # options_select = alt.selection_single(fields=['Alternative'],
    #       bind=options_dropdown, name="Alternative", empty="all"
    #   )

    frequency = (
        ax.mark_line(
            interpolate="step",
        )
        .transform_filter(
            {"not": alt.FieldEqualPredicate(field="Alternative", equal=_GENERIC)}
        )
        .encode(
            x=name,
            y="Frequency",
            color=alt.Color(
                "Alternative",
                sort=alt_names,
                legend=alt.Legend(title="Alternative", orient="left"),
            ),
            # strokeDash="Alternative",
            # tooltip="Alternative",
            opacity=alt.condition(selection, alt.value(1), alt.value(0.0)),
        )
        .add_selection(selection)
    )
    density = (
        ax.mark_line()
        .encode(
            x=name,
            y=alt.Y("Density", title="Density"),
            color=alt.Color("Alternative", legend=None, sort=alt_names),
            # strokeDash=alt.Color("Alternative", legend=None),
            # tooltip="Alternative",
            opacity=alt.condition(selection2, alt.value(1), alt.value(0.0)),
        )
        .add_selection(selection2)
        .transform_filter(selection2)
    )
    generic_density = (
        ax.mark_line()
        .encode(
            x=name,
            y=alt.Y("Density", title="Density"),
            color=alt.ColorValue("black"),
        )
        .transform_filter(
            {
                "and": [
                    alt.FieldEqualPredicate(field="Alternative", equal=_GENERIC),
                    selection,
                ]
            }
        )
    )
    density = density + generic_density

    annotation = (
        ax.mark_text(
            align="right",
            baseline="top",
            # fontSize=20,
            # dx=7
        )
        .encode(
            text="Alternative",
            x=alt.X(field=name, aggregate="max", type="quantitative", title=name),
            y=alt.Y(field="Density", aggregate="max", type="quantitative"),
        )
        .transform_filter(
            {
                "and": [
                    alt.FieldEqualPredicate(field="Alternative", equal=_GENERIC),
                    selection,
                ]
            }
        )
    )
    density = density + annotation

    if choice is not None:
        f_areas = (
            ax.mark_area(
                interpolate="step",
            )
            .encode(
                x=name,
                y=alt.Y("FrequencyChosen", title="Frequency"),
                color=alt.Color("Alternative", legend=None, sort=alt_names),
                # strokeDash="Alternative",
                # tooltip="Alternative",
                opacity=alt.condition(selection2, alt.value(0.5), alt.value(0.0)),
            )
            .transform_filter(selection2)
        )
        frequency = frequency + f_areas
        d_areas = (
            ax.mark_area()
            .encode(
                x=name,
                y=alt.Y("DensityChosen", title="Density"),
                color=alt.Color("Alternative", legend=None, sort=alt_names),
                # strokeDash="Alternative",
                # tooltip="Alternative",
                opacity=alt.condition(
                    selection2, alt.value(1 if scale_ch else 0.5), alt.value(0.0)
                ),
            )
            .transform_filter(selection2)
        )
        density = density + d_areas
        generic_d_area = (
            ax.mark_area()
            .encode(
                x=name,
                y=alt.Y("DensityChosen", title="Density"),
                opacity=alt.value(0.5),
                color=alt.ColorValue("black"),
            )
            .transform_filter(
                {
                    "and": [
                        alt.FieldEqualPredicate(field="Alternative", equal=_GENERIC),
                        selection,
                    ]
                }
            )
        )
        density = density + generic_d_area

    chart = frequency.properties(height=160, width=400) & density.properties(
        height=90, width=400
    )

    table_data = stats.to_dataframe().reset_index().reset_index()
    table_data["mean"] = table_data["mean"].apply(lambda x: f"{x:.4g}")
    table_data["std"] = table_data["std"].apply(lambda x: f"{x:.4g}")
    tab = alt.Chart(table_data)

    def make_col(colname, title):
        return (
            tab.mark_text()
            .encode(
                y=alt.Y("index:O", axis=None),
                text=colname,
                size=alt.value(10),
            )
            .properties(
                title=alt.TitleParams(text=title, align="center", fontSize=10),
            )
        )

    col1 = make_col("alt_names", "Alternative")
    col2 = make_col("mean", "Mean")
    col3 = make_col("std", "Std Dev")

    return (chart & (col1 | col2 | col3)).configure_view(strokeWidth=0), stats



# from: https://github.com/stefanholek/lazy
#
# Copyright (c) 2011-2017 Stefan H. Holek
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.


"""Decorator to create lazy attributes."""

import functools


class lazy:
    """
    Lazy descriptor.

    Used as a decorator to create lazy attributes. Lazy attributes
    are evaluated on first use.
    """

    def __init__(self, func):
        self.__func = func
        functools.wraps(self.__func)(self)

    def __get__(self, inst, inst_cls):
        if inst is None:
            return self

        if not hasattr(inst, "__dict__"):
            raise AttributeError(
                f"'{inst_cls.__name__}' object has no attribute '__dict__'"
            )

        name = self.__name__
        if name.startswith("__") and not name.endswith("__"):
            name = f"_{inst_cls.__name__}{name}"

        value = self.__func(inst)
        inst.__dict__[name] = value
        return value

    @classmethod
    def invalidate(cls, inst, name):
        """
        Invalidate a lazy attribute.

        This obviously violates the lazy contract. A subclass of lazy
        may however have a contract where invalidation is appropriate.
        """
        # These commands from the original are not currently used by larch and
        # are omitted for speed.  They are left here in case they are needed
        # for future compatibility.
        #
        # inst_cls = inst.__class__
        #
        # if not hasattr(inst, '__dict__'):
        #     raise AttributeError("'%s' object has no attribute '__dict__'" % (inst_cls.__name__,))
        #
        # if name.startswith('__') and not name.endswith('__'):
        #     name = '_%s%s' % (inst_cls.__name__, name)
        #
        # if not isinstance(getattr(inst_cls, name), cls):
        #     raise AttributeError("'%s.%s' is not a %s attribute"
        #                           % (inst_cls.__name__, name, cls.__name__))

        if name in inst.__dict__:
            del inst.__dict__[name]
